# Copyright (C) 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import itertools
import warnings
from datetime import datetime

import numpy as np
import pytest
from shapely.geometry.polygon import Polygon

from otx.api.entities.color import Color
from otx.api.entities.id import ID
from otx.api.entities.label import LabelEntity
from otx.api.entities.scored_label import Domain, ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.shapes.shape import ShapeType
from otx.api.utils.time_utils import now
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestRectangle:
    modification_date = now()

    @staticmethod
    def rectangle_labels() -> list:
        rectangle_label = LabelEntity(
            name="Rectangle label",
            domain=Domain.DETECTION,
            color=Color(red=100, green=50, blue=200),
            id=ID("rectangle_label_1"),
        )
        other_rectangle_label = LabelEntity(
            name="Other rectangle label",
            domain=Domain.SEGMENTATION,
            color=Color(red=200, green=80, blue=100),
            id=ID("rectangle_label_2"),
        )
        return [
            ScoredLabel(label=rectangle_label),
            ScoredLabel(label=other_rectangle_label),
        ]

    @staticmethod
    def horizontal_rectangle_params() -> dict:
        return {"x1": 0.1, "y1": 0.0, "x2": 0.4, "y2": 0.2}

    def horizontal_rectangle(self) -> Rectangle:
        return Rectangle(**self.horizontal_rectangle_params())

    def vertical_rectangle_params(self) -> dict:
        return {
            "x1": 0.1,
            "y1": 0.1,
            "x2": 0.3,
            "y2": 0.4,
            "modification_date": datetime(year=2020, month=1, day=1, hour=9, minute=30, second=15, microsecond=2),
        }

    def vertical_rectangle(self) -> Rectangle:
        return Rectangle(**self.vertical_rectangle_params())

    @staticmethod
    def square_params() -> dict:
        return {"x1": 0.1, "y1": 0.1, "x2": 0.3, "y2": 0.3}

    def square(self) -> Rectangle:
        return Rectangle(**self.square_params())

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_required_parameters(self):
        """
        <b>Description:</b>
        Check Rectangle class instance required parameters

        <b>Input data:</b>
        Rectangle class initiation parameters

        <b>Expected results:</b>
        Test passes if Rectangle instance has expected attributes specified during required parameters initiation

        <b>Steps</b>
        1. Compare x1, y1, x2, y2 and type Rectangle instance parameters with expected values
        """
        rectangle = self.horizontal_rectangle()
        assert rectangle.x1 == 0.1
        assert rectangle.y1 == 0.0
        assert rectangle.x2 == 0.4
        assert rectangle.y2 == 0.2
        assert rectangle.type == ShapeType.RECTANGLE

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_optional_parameters(self):
        """
        <b>Description:</b>
        Check Rectangle optional parameters

        <b>Input data:</b>
        Instance of Rectangle class

        <b>Expected results:</b>
        Test passes if Rectangle instance has expected modification attributes specified during Rectangle
        class object initiation with optional parameters

        <b>Steps</b>
        1. Compare default Rectangle instance attribute with expected value
        2. Check type of default modification_date Rectangle instance attribute
        3. Compare specified Rectangle instance attribute with expected value
        4. Compare specified modification_date Rectangle instance attribute with expected value
        """
        # Checking default values of optional parameters
        default_params_rectangle = self.horizontal_rectangle()
        assert isinstance(default_params_rectangle.modification_date, datetime)
        # check for specified values of optional parameters
        specified_params_rectangle = self.vertical_rectangle()
        assert specified_params_rectangle.modification_date == datetime(
            year=2020, month=1, day=1, hour=9, minute=30, second=15, microsecond=2
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_coordinates_validation(self):
        """
        <b>Description:</b>
        Check Rectangle coordinates

        <b>Input data:</b>
        Rectangle class initiation parameters

        <b>Expected results:</b>
        Test passes if Rectangle correctly checks coordinates during initiation of Rectangle object

        <b>Steps</b>
        1. Check Rectangle coordinates with correct width and height
        2. Check Rectangle coordinates with incorrect width: (x2 - x1) <= 0
        3. Check Rectangle coordinates with incorrect height: (y2 - y1) <= 0
        4. Check Rectangle coordinates with all coordinates equal 0
        """
        # checks for correct width and height
        self.horizontal_rectangle()
        self.vertical_rectangle()
        self.square()
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "Rectangle coordinates")
            Rectangle(x1=0.2, y1=0.1, x2=1.4, y2=1.5)
        Rectangle(x1=0.2, y1=0.1, x2=0.4, y2=0.5)
        # checks for incorrect coordinates
        width_less_than_zero_params = {"x1": 0.4, "y1": 0.0, "x2": 0.1, "y2": 0.2}
        width_equal_zero_params = {"x1": 0.1, "y1": 0.0, "x2": 0.1, "y2": 0.2}
        height_less_than_zero_params = {"x1": 0.1, "y1": 0.4, "x2": 0.3, "y2": 0.1}
        height_params_equal_zero_params = {"x1": 0.1, "y1": 0.4, "x2": 0.3, "y2": 0.4}
        zero_rectangle_params = {"x1": 0.0, "x2": 0.0, "y1": 0.0, "y2": 0.0}
        for incorrect_coordinates in [
            width_less_than_zero_params,
            width_equal_zero_params,
            height_less_than_zero_params,
            height_params_equal_zero_params,
            zero_rectangle_params,
        ]:
            with pytest.raises(ValueError):
                Rectangle(**incorrect_coordinates)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_repr(self):
        """
        <b>Description:</b>
        Check Rectangle __repr__ method

        <b>Input data:</b>
        Instance of Rectangle class

        <b>Expected results:</b>
        Test passes if __repr__ method prints expected message for Rectangle class instance

        <b>Steps</b>
        1. Check message printed by __repr__ method
        """
        rectangle = self.horizontal_rectangle()
        assert repr(rectangle) == "Rectangle(x=0.1, y=0.0, width=0.30000000000000004, height=0.2)"

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_eq(self):
        """
        <b>Description:</b>
        Check Rectangle __eq__ method

        <b>Input data:</b>
        Instances of Rectangle class

        <b>Expected results:</b>
        Test passes if __eq__ method returns expected bool value

        <b>Steps</b>
        1. Check __eq__ method for instances of Rectangle class with equal parameters
        2. Check __eq__ method for different instances of Rectangle class
        3. Check __eq__ method for instances of different classes
        4. Check __eq__ method for instances of Rectangle class with unequal x1, y1, x2, y2 and
        modification_date attributes
        """
        rectangle = self.vertical_rectangle()
        # Check for instances with equal parameters
        equal_rectangle = self.vertical_rectangle()
        assert rectangle == equal_rectangle
        # Check for different instances of Rectangle class
        assert rectangle != self.horizontal_rectangle()
        # Check for different types branch
        assert rectangle != str

        assert rectangle == equal_rectangle
        # Check for instances with unequal parameters combinations
        # Generating all possible scenarios of parameter values submission
        keys_list = ["x1", "y1", "x2", "y2", "modification_date"]
        parameter_combinations = []
        for i in range(1, len(keys_list) + 1):
            parameter_combinations.append(list(itertools.combinations(keys_list, i)))
        # In each of scenario creating a copy of equal Rectangle parameters and replacing to values from prepared
        # dictionary
        unequal_values_dict = {
            "x1": 0.09,
            "y1": 0.09,
            "x2": 0.29,
            "y2": 0.39,
            "modification_date": datetime(year=2019, month=2, day=3, hour=7, minute=15, second=10, microsecond=1),
        }
        for scenario in parameter_combinations:
            for rectangle_parameters in scenario:
                unequal_rectangle_params_dict = dict(self.vertical_rectangle_params())
                for key in rectangle_parameters:
                    unequal_rectangle_params_dict[key] = unequal_values_dict.get(key)
                unequal_rectangle = Rectangle(**unequal_rectangle_params_dict)
                assert rectangle != unequal_rectangle, (
                    "Failed to check that Rectangle instances with different " f"{rectangle_parameters} are unequal"
                )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_hash(self):
        """
        <b>Description:</b>
        Check Rectangle __hash__ method

        <b>Input data:</b>
        Instance of Rectangle class

        <b>Expected results:</b>
        Test passes if __hash__ method returns expected value

        <b>Steps</b>
        1. Check value returned by __hash__ method
        """
        rectangle = self.horizontal_rectangle()
        assert hash(rectangle) == hash(str(rectangle))

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_clip_to_visible_region(self):
        """
        <b>Description:</b>
        Check Rectangle clip_to_visible_region method

        <b>Input data:</b>
        Rectangle class initiation parameters

        <b>Expected results:</b>
        Test passes if clip_to_visible_region method return correct value

        <b>Steps</b>
        1. Check values returned by clip_to_visible_region method for 0<x1<x2, 0<y1<y2, x1<x2<1, y1<y2<1
        2. Check values returned by clip_to_visible_region method for x1<0, y1<0, x2>1, y2>1
        3. Check values returned by clip_to_visible_region method for x1=0, y1=0, x2=1, y2=1
        4. Check ValueError exception raised if x1<0 and x1<x2<0: clipped Rectangle width will be equal 0
        5. Check ValueError exception raised if 1<x1<x2 and x2>1: clipped Rectangle width will be equal 0
        6. Check ValueError exception raised if y1<0 and y1<y2<0: clipped Rectangle height will be equal 0
        7. Check ValueError exception raised if 1<y1<y2 and y2>1: clipped Rectangle height will be equal 0
        """
        positive_scenarios = [
            {
                "input_params": {
                    "x1": 0.3,
                    "y1": 0.2,
                    "x2": 0.6,
                    "y2": 0.4,
                },
                "params_expected": {"x1": 0.3, "y1": 0.2, "x2": 0.6, "y2": 0.4},
            },
            {
                "input_params": {
                    "x1": -0.2,
                    "y1": -0.3,
                    "x2": 1.6,
                    "y2": 1.4,
                },
                "params_expected": {"x1": 0.0, "y1": 0.0, "x2": 1.0, "y2": 1.0},
            },
            {
                "input_params": {
                    "x1": 0.0,
                    "y1": 0.0,
                    "x2": 1.0,
                    "y2": 1.0,
                },
                "params_expected": {"x1": 0.0, "y1": 0.0, "x2": 1.0, "y2": 1.0},
            },
        ]
        for scenario in positive_scenarios:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", "Rectangle coordinates")
                rectangle_actual = Rectangle(**scenario.get("input_params"))
            rectangle_expected = Rectangle(**scenario.get("params_expected"))
            rectangle_actual.modification_date = self.modification_date
            rectangle_expected.modification_date = self.modification_date
            assert rectangle_actual.clip_to_visible_region() == rectangle_expected
        negative_scenarios = [
            {"x1": -0.4, "y1": 0.2, "x2": -0.2, "y2": 0.4},
            {"x1": 1.2, "y1": 0.2, "x2": 1.6, "y2": 0.4},
            {"x1": 0.4, "y1": -0.4, "x2": 0.6, "y2": -0.2},
            {"x1": 1.2, "y1": 1.2, "x2": 1.6, "y2": 1.4},
        ]
        for scenario in negative_scenarios:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", "Rectangle coordinates")
                rectangle_actual = Rectangle(**scenario)
            with pytest.raises(ValueError):
                rectangle_actual.clip_to_visible_region()

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_normalize_wrt_roi_shape(self):
        """
        <b>Description:</b>
        Check Rectangle normalize_wrt_roi_shape method

        <b>Input data:</b>
        Instance of Rectangle class

        <b>Expected results:</b>
        Test passes if normalize_wrt_roi_shape method returns expected instance of Rectangle class

        <b>Steps</b>
        1. Check values returned by normalized Rectangle instance
        2. Check raise ValueError exception when roi parameter has unexpected type
        """
        # Positive scenario
        rectangle = self.horizontal_rectangle()
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "Rectangle coordinates")
            roi_shape = Rectangle(x1=0.0, y1=0.0, x2=2.1, y2=2.2)
        normalized = rectangle.normalize_wrt_roi_shape(roi_shape)
        assert normalized.x1 == 0.1
        assert normalized.y1 == 0.0
        assert normalized.x2 == 0.4
        assert normalized.y2 == 0.2
        assert normalized.modification_date == rectangle.modification_date
        # Negative scenario
        with pytest.raises(ValueError):
            rectangle.normalize_wrt_roi_shape(str)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_denormalize_wrt_roi_shape(self):
        """
        <b>Description:</b>
        Check Rectangle denormalize_wrt_roi_shape method

        <b>Input data:</b>
        Instance of Rectangle class

        <b>Expected results:</b>
        Test passes if denormalize_wrt_roi_shape method return expected instance of Rectangle class

        <b>Steps</b>
        1. Check values returned by denormalized Rectangle instance
        2. Check raise ValueError exception when roi parameter has unexpected type
        """
        # Positive scenario
        rectangle = self.horizontal_rectangle()
        roi_shape = Rectangle(x1=0.2, y1=0.2, x2=0.4, y2=0.4)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "Rectangle coordinates")
            denormalized = rectangle.denormalize_wrt_roi_shape(roi_shape)
        assert denormalized.x1 == -0.5
        assert denormalized.y1 == -1.0
        assert denormalized.x2 == 1.0
        assert denormalized.y2 == 0.0
        assert denormalized.modification_date == rectangle.modification_date
        # Negative scenario
        with pytest.raises(ValueError):
            rectangle.denormalize_wrt_roi_shape(str)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_as_shapely_polygon(self):
        """
        <b>Description:</b>
        Check Rectangle _as_shapely_polygon method

        <b>Input data:</b>
        Instance of Rectangle class

        <b>Expected results:</b>
        Test passes if _as_shapely_polygon method returns expected instance of Polygon class

        <b>Steps</b>
        1. Check Polygon instance returned by _as_shapely_polygon method
        """
        rectangle = self.horizontal_rectangle()
        shapely_polygon = rectangle._as_shapely_polygon()
        assert shapely_polygon.__class__ == Polygon
        assert shapely_polygon.bounds == (0.1, 0.0, 0.4, 0.2)
        assert shapely_polygon.area == 0.06000000000000001

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_generate_full_box(self):
        """
        <b>Description:</b>
        Check Rectangle generate_full_box method

        <b>Expected results:</b>
        Test passes if generate_full_box method returns instance of Rectangle class with coordinates
        (x1=0.0, y1=0.0, x2=1.0, y2=1.0)

        <b>Steps</b>
        1. Check generate_full_box method for Rectangle instance
        """
        full_box = Rectangle.generate_full_box()
        assert full_box.type == ShapeType.RECTANGLE
        assert full_box.x1 == full_box.y1 == 0.0
        assert full_box.x2 == full_box.y2 == 1.0

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_is_full_box(self):
        """
        <b>Description:</b>
        Check Rectangle is_full_box method

        <b>Input data:</b>
        Rectangle class initiation parameters

        <b>Expected results:</b>
        Test passes if is_full_box method return expected bool value

        <b>Steps</b>
        1. Check positive scenario for is_full_box method: x1=y1=0, x2=y2=1
        2. Check negative scenarios for is_full_box method
        """
        full_box_rectangle_params = {"x1": 0.0, "y1": 0.0, "x2": 1.0, "y2": 1.0}
        # Positive scenario for Rectangle instance
        full_box_rectangle = Rectangle(**full_box_rectangle_params)
        assert Rectangle.is_full_box(full_box_rectangle)
        # Negative scenarios for Rectangle instance
        # Generating all scenarios when Rectangle object not a full_box
        keys_list = ["x1", "y1", "x2", "y2"]
        parameter_combinations = []
        for i in range(1, len(keys_list) + 1):
            parameter_combinations.append(list(itertools.combinations(keys_list, i)))
        # In each of scenario creating a copy of full_Box rectangle parameters and replacing to values from prepared
        # dictionary
        not_full_box_values_dict = {"x1": 0.01, "y1": 0.01, "x2": 0.9, "y2": 0.9}
        for combination in parameter_combinations:
            for scenario in combination:
                not_full_box_params = dict(full_box_rectangle_params)
                for key in scenario:
                    not_full_box_params[key] = not_full_box_values_dict.get(key)
                not_full_box_rectangle = Rectangle(**not_full_box_params)
                assert not Rectangle.is_full_box(not_full_box_rectangle), (
                    f"Expected False returned by is_full_box method for rectangle with parameters "
                    f"{not_full_box_params}"
                )
        # Negative scenario for not Rectangle class instance
        assert not Rectangle.is_full_box(str)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_crop_numpy_array(self):
        """
        <b>Description:</b>
        Check Rectangle crop_numpy_array method

        <b>Input data:</b>
        Rectangle class initiation parameters

        <b>Expected results:</b>
        Test passes if crop_numpy_array method return expected cropped array
        <b>Steps</b>
        1. Check crop_numpy_array method for Rectangle with parameters less than 0
        2. Check crop_numpy_array method for Rectangle with parameters range from 0 to 1
        3. Check crop_numpy_array method for Rectangle with parameters more than 1
        """
        image_height = image_width = 128
        numpy_image_array = np.random.uniform(low=0.0, high=255.0, size=(image_height, image_width, 3))
        scenarios = [
            {
                "input_params": {"x1": -0.2, "x2": -0.1, "y1": -0.3, "y2": -0.2},
                "cropped_expected": {"x1": 0, "y1": 0, "x2": 0, "y2": 0},
            },
            {
                "input_params": {"x1": 0.2, "x2": 0.3, "y1": 0.4, "y2": 0.8},
                "cropped_expected": {"x1": 26, "y1": 51, "x2": 38, "y2": 102},
            },
            {
                "input_params": {"x1": 1.1, "x2": 1.3, "y1": 1.1, "y2": 1.5},
                "cropped_expected": {"x1": 141, "y1": 141, "x2": 166, "y2": 192},
            },
        ]
        for rectangle_parameters in scenarios:
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", "Rectangle coordinates")
                rectangle = Rectangle(**rectangle_parameters.get("input_params"))
            expected_output = rectangle_parameters.get("cropped_expected")
            actual_cropped_image_array = rectangle.crop_numpy_array(numpy_image_array)
            expected_image_array = numpy_image_array[
                expected_output.get("y1") : expected_output.get("y2"),
                expected_output.get("x1") : expected_output.get("x2"),
                ::,
            ]
            assert actual_cropped_image_array.shape[2] == 3
            try:
                assert (expected_image_array == actual_cropped_image_array).all()
            except AttributeError:
                raise AssertionError("Unequal expected and cropped image arrays")

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_width(self):
        """
        <b>Description:</b>
        Check Rectangle width method

        <b>Input data:</b>
        Instances of Rectangle class

        <b>Expected results:</b>
        Test passes if width method returns expected value of Rectangle width

        <b>Steps</b>
        1. Check width method for Rectangle instances
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "Rectangle coordinates")
            negative_x1_rectangle = Rectangle(x1=-0.3, y1=0.2, x2=0.7, y2=0.5)
        for rectangle, expected_width in [
            (self.horizontal_rectangle(), 0.30000000000000004),
            (self.vertical_rectangle(), 0.19999999999999998),
            (self.square(), 0.19999999999999998),
            (negative_x1_rectangle, 1.0),
        ]:
            assert rectangle.width == expected_width

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_height(self):
        """
        <b>Description:</b>
        Check Rectangle height method

        <b>Input data:</b>
        Instances of Rectangle class

        <b>Expected results:</b>
        Test passes if height method returns expected value of Rectangle height

        <b>Steps</b>
        1. Check height method for Rectangle instances
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "Rectangle coordinates")
            negative_y1_rectangle = Rectangle(x1=0.3, y1=-0.4, x2=0.7, y2=0.5)
        for rectangle, expected_height in [
            (self.horizontal_rectangle(), 0.2),
            (self.vertical_rectangle(), 0.30000000000000004),
            (self.square(), 0.19999999999999998),
            (negative_y1_rectangle, 0.9),
        ]:
            assert rectangle.height == expected_height

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_diagonal(self):
        """
        <b>Description:</b>
        Check Rectangle diagonal method

        <b>Input data:</b>
        Instances of Rectangle class

        <b>Expected results:</b>
        Test passes if diagonal method returns expected value of Rectangle diagonal

        <b>Steps</b>
        1. Check diagonal method for Rectangle instance
        """
        for rectangle, expected_diagonal in [
            (self.horizontal_rectangle(), 0.36055512754639896),
            (self.vertical_rectangle(), 0.36055512754639896),
            (self.square(), 0.282842712474619),
        ]:
            np.testing.assert_approx_equal(rectangle.diagonal, expected_diagonal)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_rectangle_get_area(self):
        """
        <b>Description:</b>
        Check Rectangle get_area method

        <b>Input data:</b>
        Instances of Rectangle class

        <b>Expected results:</b>
        Test passes if get_area method returns expected value of Rectangle diagonal

        <b>Steps</b>
        1. Check get_area method for Rectangle instance
        """
        for rectangle, expected_area in [
            (self.horizontal_rectangle(), 0.06000000000000001),
            (self.vertical_rectangle(), 0.060000000000000005),
            (self.square(), 0.039999999999999994),
        ]:
            assert rectangle.get_area() == expected_area



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-Lggggcense-Identifier: Apache-2.0
#

import pytest
import torch
from torch.nn import functional as F

from otx.core.ov.ops.poolings import AvgPoolV1, MaxPoolV0
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestMaxPoolV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 3, 256, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_invalid_attr(self):
        with pytest.raises(ValueError):
            MaxPoolV0(
                "dummy",
                shape=self.shape,
                strides=[1, 1],
                pads_begin=[0, 0],
                pads_end=[0, 0],
                kernel=[8, 8],
                rounding_type="error",
            )

        with pytest.raises(ValueError):
            MaxPoolV0(
                "dummy",
                shape=self.shape,
                strides=[1, 1],
                pads_begin=[0, 0],
                pads_end=[0, 0],
                kernel=[8, 8],
                auto_pad="error",
            )

        with pytest.raises(ValueError):
            MaxPoolV0(
                "dummy",
                shape=self.shape,
                strides=[1, 1],
                pads_begin=[0, 0],
                pads_end=[0, 0],
                kernel=[8, 8],
                index_element_type="error",
            )

        with pytest.raises(NotImplementedError):
            MaxPoolV0(
                "dummy",
                shape=self.shape,
                strides=[1, 1],
                pads_begin=[0, 0],
                pads_end=[0, 0],
                kernel=[8, 8],
                axis=1,
            )

    @e2e_pytest_unit
    def test_forward(self):
        op = MaxPoolV0(
            "dummy",
            shape=self.shape,
            strides=[1, 1],
            pads_begin=[0, 0],
            pads_end=[0, 0],
            kernel=[8, 8],
        )

        with pytest.raises(NotImplementedError):
            op(torch.randn(1, 1, 1, 1, 1, 1))

        output = op(self.input)
        ref = F.max_pool2d(
            self.input,
            op.attrs.kernel,
            op.attrs.strides,
            dilation=op.attrs.dilations,
            ceil_mode=op.attrs.rounding_type == "ceil",
            return_indices=True,
        )
        for i, j in zip(output, ref):
            assert torch.equal(i, j)


class TestAvgPoolV1:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 3, 256, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_invalid_attr(self):
        with pytest.raises(ValueError):
            AvgPoolV1(
                "dummy",
                shape=self.shape,
                exclude_pad=False,
                strides=[1, 1],
                pads_begin=[0, 0],
                pads_end=[0, 0],
                kernel=[8, 8],
                rounding_type="error",
            )

        with pytest.raises(ValueError):
            AvgPoolV1(
                "dummy",
                shape=self.shape,
                exclude_pad=False,
                strides=[1, 1],
                pads_begin=[0, 0],
                pads_end=[0, 0],
                kernel=[8, 8],
                auto_pad="error",
            )

    @e2e_pytest_unit
    def test_forward(self):
        op = AvgPoolV1(
            "dummy",
            shape=self.shape,
            exclude_pad=False,
            strides=[1, 1],
            pads_begin=[0, 0],
            pads_end=[0, 0],
            kernel=[8, 8],
        )

        with pytest.raises(NotImplementedError):
            op(torch.randn(1, 1, 1, 1, 1, 1))

        output = op(self.input)
        ref = F.avg_pool2d(
            self.input,
            op.attrs.kernel,
            op.attrs.strides,
            ceil_mode=op.attrs.rounding_type == "ceil",
            count_include_pad=not op.attrs.exclude_pad,
        )
        assert torch.equal(output, ref)



"""Unit-Test case for otx.core.data.adapter.segmentation_dataset_adapter."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import os
import shutil
from typing import Optional

import numpy as np
import pytest

from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.model_template import TaskType
from otx.api.entities.subset import Subset
from otx.core.data.adapter.segmentation_dataset_adapter import (
    SegmentationDatasetAdapter,
    SelfSLSegmentationDatasetAdapter,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.core.data.test_helpers import (
    TASK_NAME_TO_DATA_ROOT,
    TASK_NAME_TO_TASK_TYPE,
)


class TestOTXSegmentationDatasetAdapter:
    def setup_method(self):
        self.root_path = os.getcwd()
        task = "segmentation"

        self.task_type: TaskType = TASK_NAME_TO_TASK_TYPE[task]
        data_root_dict: dict = TASK_NAME_TO_DATA_ROOT[task]

        self.train_data_roots: str = os.path.join(self.root_path, data_root_dict["train"])
        self.val_data_roots: str = os.path.join(self.root_path, data_root_dict["val"])
        self.test_data_roots: str = os.path.join(self.root_path, data_root_dict["test"])
        self.unlabeled_data_roots: Optional[str] = None
        if "unlabeled" in data_root_dict:
            self.unlabeled_data_roots = os.path.join(self.root_path, data_root_dict["unlabeled"])

        self.train_dataset_adapter = SegmentationDatasetAdapter(
            task_type=self.task_type,
            train_data_roots=self.train_data_roots,
            val_data_roots=self.val_data_roots,
            unlabeled_data_roots=self.unlabeled_data_roots,
        )

        self.test_dataset_adapter = SegmentationDatasetAdapter(
            task_type=self.task_type,
            test_data_roots=self.test_data_roots,
        )

    @e2e_pytest_unit
    def test_init(self):
        assert Subset.TRAINING in self.train_dataset_adapter.dataset
        assert Subset.VALIDATION in self.train_dataset_adapter.dataset
        if self.unlabeled_data_roots is not None:
            assert Subset.UNLABELED in self.train_dataset_adapter.dataset

        assert Subset.TESTING in self.test_dataset_adapter.dataset

    @e2e_pytest_unit
    def test_get_otx_dataset(self):
        assert isinstance(self.train_dataset_adapter.get_otx_dataset(), DatasetEntity)
        assert isinstance(self.test_dataset_adapter.get_otx_dataset(), DatasetEntity)


class TestSelfSLSegmentationDatasetAdapter:
    def setup_method(self, method) -> None:
        self.root_path = os.getcwd()
        task = "segmentation"

        self.task_type: TaskType = TASK_NAME_TO_TASK_TYPE[task]
        data_root_dict: dict = TASK_NAME_TO_DATA_ROOT[task]
        self.train_data_roots: str = os.path.join(self.root_path, data_root_dict["train"], "images")

        self.pseudo_mask_roots = os.path.abspath(self.train_data_roots.replace("images", "detcon_mask"))

    @e2e_pytest_unit
    def test_import_dataset_create_all_masks(self, mocker):
        """Test _import_dataset when creating all masks.

        This test is for when all masks are not created and it is required to create masks.
        """
        shutil.rmtree(self.pseudo_mask_roots, ignore_errors=True)
        spy_create_pseudo_masks = mocker.spy(SelfSLSegmentationDatasetAdapter, "create_pseudo_masks")

        dataset_adapter = SelfSLSegmentationDatasetAdapter(
            task_type=self.task_type,
            train_data_roots=self.train_data_roots,
        )

        spy_create_pseudo_masks.assert_called()
        assert spy_create_pseudo_masks.call_count == len(dataset_adapter.dataset[Subset.TRAINING])

    @e2e_pytest_unit
    @pytest.mark.parametrize("idx_remove", [1, 2, 3])
    def test_import_dataset_create_some_uncreated_masks(self, mocker, idx_remove: int):
        """Test _import_dataset when there are both uncreated and created masks.

        This test is for when there are both created and uncreated masks
        and it is required to either create or just load masks.
        In this test, remove a mask created before and check if `create_pseudo_masks` is called once.
        """
        shutil.rmtree(self.pseudo_mask_roots, ignore_errors=True)
        dataset_adapter = SelfSLSegmentationDatasetAdapter(
            task_type=self.task_type,
            train_data_roots=self.train_data_roots,
        )
        assert os.path.isdir(self.pseudo_mask_roots)
        assert len(os.listdir(self.pseudo_mask_roots)) == 4

        # remove a mask
        os.remove(os.path.join(self.pseudo_mask_roots, f"000{idx_remove}.png"))
        spy_create_pseudo_masks = mocker.spy(SelfSLSegmentationDatasetAdapter, "create_pseudo_masks")

        _ = dataset_adapter._import_dataset(
            train_data_roots=self.train_data_roots,
        )

        spy_create_pseudo_masks.assert_called()
        assert spy_create_pseudo_masks.call_count == 1

    @e2e_pytest_unit
    def test_import_dataset_just_load_masks(self, mocker):
        """Test _import_dataset when just loading all masks."""
        spy_create_pseudo_masks = mocker.spy(SelfSLSegmentationDatasetAdapter, "create_pseudo_masks")

        _ = SelfSLSegmentationDatasetAdapter(
            task_type=self.task_type,
            train_data_roots=self.train_data_roots,
        )

        spy_create_pseudo_masks.assert_not_called()

    @e2e_pytest_unit
    @pytest.mark.xfail
    def test_get_otx_dataset_without_skipping_background(self):
        """Test get_otx_dataset without skipping background.

        TODO (sungchul): don't skip background class in get_otx_dataset
        """
        assert 0

    @e2e_pytest_unit
    def test_create_pseudo_masks(self, mocker):
        """Test create_pseudo_masks."""
        mocker.patch("otx.core.data.adapter.segmentation_dataset_adapter.os.makedirs")
        mocker.patch("otx.core.data.adapter.segmentation_dataset_adapter.cv2.imwrite")
        dataset_adapter = SelfSLSegmentationDatasetAdapter(
            task_type=self.task_type,
            train_data_roots=self.train_data_roots,
        )

        pseudo_mask = dataset_adapter.create_pseudo_masks(img=np.ones((2, 2)), pseudo_mask_path="")

        assert type(pseudo_mask) == np.ndarray



"""This module tests classes related to image"""

# Copyright (C) 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


import os
import tempfile

import cv2
import numpy as np
import pytest

from otx.api.entities.annotation import Annotation
from otx.api.entities.image import Image
from otx.api.entities.shapes.ellipse import Ellipse
from otx.api.entities.shapes.rectangle import Rectangle
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestImage:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_image(self):
        """
        <b>Description:</b>
        To test Image class

        <b>Input data:</b>
        Instances of Image sourced with data or file path to forged image or both
        Rectangle and Ellipse classes instances
        Instances of Annotation made against rectangle and Ellipsis

        <b>Expected results:</b>
        1. Value error is raised
        2. Value error is raised
        3. Created successfully
        4. Created successfully
        5. Calls of the method against both instances return expected strings
        6. Numpy property contains expected data
        7. Numpy property is changed and contains expected data
        8. Called method returns expected value
        9. Value error raises
        10. Value error raises
        11. Called method returns expected value
        12. Height property contains expected data
        13. Width property contains expected data

        <b>Steps</b>
        1. Create Image without parameters
        2. Create Image with both parameters
        3. Create Image with data parameter
        4. Create Image with path to forged image
        5. Check __str__ method against data instance and fp instance
        6. Check numpy property against data instance and fp instance
        7. Change numpy property and check it against data instance and fp instance
        8. Check roi_numpy method against data instance and fp instance without Annotation
        9. Check roi_numpy method against data instance and fp instance with Annotation of Ellipsis shape
        10. Check roi_numpy method against 1 dimensional data instance with Annotation of Rectangle shape
        11. Check roi_numpy property against data instance and fp instance with Annotation of Rectangle shape
        12. Check height property against data instance and fp instance
        13. Check width property against data instance and fp instance
        """
        test_height = test_width = 128
        test_height1 = test_width1 = 64
        test_depth = 4
        data0 = np.random.random_sample((test_height, test_width, test_depth))
        data1 = np.random.random_sample((test_height1, test_width1, test_depth))
        d1_data = np.random.random_sample((test_height1,))
        image = np.random.random_sample((test_height, test_width, test_depth))
        image_path = os.path.join(tempfile.gettempdir(), "test_image.png")
        cv2.imwrite(image_path, image)

        with pytest.raises(ValueError):
            Image()

        with pytest.raises(ValueError):
            Image(data=data0, file_path=image_path)

        data_instance = Image(data=data0)
        assert isinstance(data_instance, Image)

        fp_instance = Image(file_path=image_path)
        assert isinstance(fp_instance, Image)

        assert str(data_instance) == f"Image(with data, width={test_width}, height={test_height})"
        assert str(fp_instance) == f"Image({image_path}, width={test_width}, height={test_height})"

        assert np.array_equal(data_instance.numpy, data0)
        height, width, depth = fp_instance.numpy.shape
        assert height == test_height
        assert width == test_width
        assert depth != test_depth

        data_instance.numpy = data1
        fp_instance.numpy = data1

        assert np.array_equal(data_instance.numpy, data1)
        assert np.array_equal(fp_instance.numpy, data1)

        assert np.array_equal(data_instance.roi_numpy(), data1)
        assert np.array_equal(fp_instance.roi_numpy(), data1)

        rec_shape = Rectangle(x1=0.0, x2=0.5, y1=0.0, y2=0.5)
        ellipsis_shape = Ellipse(x1=0.0, x2=0.5, y1=0.0, y2=0.5)
        label = []
        rect_annotation = Annotation(rec_shape, label)
        ellipsis_annotation = Annotation(ellipsis_shape, label)

        with pytest.raises(ValueError):
            data_instance.roi_numpy(ellipsis_annotation)

        with pytest.raises(ValueError):
            fp_instance.roi_numpy(ellipsis_annotation)

        d1_data_instance = Image(data=d1_data)

        with pytest.raises(ValueError):
            d1_data_instance.roi_numpy(rect_annotation)

        for arr in data_instance.roi_numpy(rect_annotation):
            for it in arr:
                assert it in data1

        assert data_instance.height == fp_instance.height == test_height1
        assert data_instance.width == fp_instance.width == test_width1

        if os.path.exists(image_path):
            os.remove(image_path)



import pytest

from otx.algorithms.common.adapters.torch.utils import adapt_batch_size
from otx.algorithms.common.adapters.torch.utils import bs_search_algo


def test_adapt_batch_size(mocker):
    mocker_torch = mocker.patch.object(bs_search_algo, "torch")
    mocker_torch.cuda.mem_get_info.return_value = (1, 10000)

    def mock_train_func(batch_size):
        if batch_size > 100:
            mocker_torch.cuda.max_memory_allocated.return_value = 10000
            raise RuntimeError("CUDA out of memory.")
        elif batch_size > 80:
            mocker_torch.cuda.max_memory_allocated.return_value = 9000
        else:
            mocker_torch.cuda.max_memory_allocated.return_value = 1000

    adapted_bs = adapt_batch_size(mock_train_func, 128, 1000)

    assert adapted_bs == 80


def test_adapt_batch_size_gpu_memory_too_small(mocker):
    mocker_torch = mocker.patch.object(bs_search_algo, "torch")
    mocker_torch.cuda.mem_get_info.return_value = (1, 10000)

    def mock_train_func(batch_size):
        if batch_size > 4:
            mocker_torch.cuda.max_memory_allocated.return_value = 10000
            raise RuntimeError("CUDA out of memory.")
        elif batch_size >= 2:
            mocker_torch.cuda.max_memory_allocated.return_value = 9000
        else:
            mocker_torch.cuda.max_memory_allocated.return_value = 1000

    with pytest.raises(RuntimeError):
        adapt_batch_size(mock_train_func, 128, 1000)


@pytest.mark.parametrize("default_bs", [-1, 0])
def test_adapt_batch_size_wrong_default_bs(mocker, default_bs):
    with pytest.raises(ValueError):
        adapt_batch_size(mocker.MagicMock(), default_bs, 1000)


@pytest.mark.parametrize("trainset_size", [-1, 0])
def test_adapt_batch_size_wrong_trainset_size(mocker, trainset_size):
    with pytest.raises(ValueError):
        adapt_batch_size(mocker.MagicMock(), 8, trainset_size)



"""Test Helpers for otx.core.data."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
from typing import List, Optional
import os

import cv2
import datumaro as dm
import numpy as np

from otx.api.entities.model_template import TaskType

TASK_NAME_TO_TASK_TYPE = {
    "classification": TaskType.CLASSIFICATION,
    "detection": TaskType.DETECTION,
    "rotated_detection": TaskType.ROTATED_DETECTION,
    "instance_segmentation": TaskType.INSTANCE_SEGMENTATION,
    "segmentation": TaskType.SEGMENTATION,
    "anomaly_classification": TaskType.ANOMALY_CLASSIFICATION,
    "anomaly_detection": TaskType.ANOMALY_DETECTION,
    "anomaly_segmentation": TaskType.ANOMALY_SEGMENTATION,
    "action_classification": TaskType.ACTION_CLASSIFICATION,
    "action_detection": TaskType.ACTION_DETECTION,
}

TASK_NAME_TO_DATA_ROOT = {
    "classification": {
        "train": "tests/assets/classification_dataset",
        "val": "tests/assets/classification_dataset",
        "test": "tests/assets/classification_dataset",
        "unlabeled": "tests/assets/classification_dataset",
    },
    "detection": {
        "train": "tests/assets/car_tree_bug",
        "val": "tests/assets/car_tree_bug",
        "test": "tests/assets/car_tree_bug",
        "unlabeled": "tests/assets/car_tree_bug",
    },
    "rotated_detection": {
        "train": "tests/assets/car_tree_bug",
        "val": "tests/assets/car_tree_bug",
        "test": "tests/assets/car_tree_bug",
        "unlabeled": "tests/assets/car_tree_bug",
    },
    "instance_segmentation": {
        "train": "tests/assets/car_tree_bug",
        "val": "tests/assets/car_tree_bug",
        "test": "tests/assets/car_tree_bug",
    },
    "segmentation": {
        "train": "tests/assets/common_semantic_segmentation_dataset/train",
        "val": "tests/assets/common_semantic_segmentation_dataset/val",
        "test": "tests/assets/common_semantic_segmentation_dataset/val",
        "unlabeled": "tests/assets/common_semantic_segmentation_dataset/val",
    },
    "anomaly_classification": {
        "train": "tests/assets/anomaly/hazelnut",
        "val": "tests/assets/anomaly/hazelnut",
        "test": "tests/assets/anomaly/hazelnut",
    },
    "anomaly_detection": {
        "train": "tests/assets/anomaly/hazelnut",
        "val": "tests/assets/anomaly/hazelnut",
        "test": "tests/assets/anomaly/hazelnut",
    },
    "anomaly_segmentation": {
        "train": "tests/assets/anomaly/hazelnut",
        "val": "tests/assets/anomaly/hazelnut",
        "test": "tests/assets/anomaly/hazelnut",
    },
    "action_classification": {
        "train": "tests/assets/cvat_dataset/action_classification/train",
        "val": "tests/assets/cvat_dataset/action_classification/train",
        "test": "tests/assets/cvat_dataset/action_classification/train",
    },
    "action_detection": {
        "train": "tests/assets/cvat_dataset/action_detection/train",
        "val": "tests/assets/cvat_dataset/action_detection/train",
        "test": "tests/assets/cvat_dataset/action_detection/train",
    },
}


def generate_datumaro_dataset_item(
    item_id: str,
    subset: str,
    task: str,
    image_shape: np.array = np.array((5, 5, 3)),
    mask_shape: np.array = np.array((5, 5)),
    temp_dir: Optional[str] = None,
) -> dm.DatasetItem:
    """Generate Datumaro DatasetItem.

    Args:
        item_id (str): The ID of dataset item
        subset (str): subset of item, e.g. "train" or "val"
        task (str): task type, e.g. "classification"
        image_shape (np.array): the shape of image.
        image_shape (np.array): the shape of mask.
        temp_dir (str): directory to save image data

    Returns:
        dm.DatasetItem: Datumaro DatasetItem
    """
    ann_task_dict = {
        "classification": dm.Label(label=0),
        "detection": dm.Bbox(1, 2, 3, 4, label=0),
        "segmentation": dm.Mask(np.zeros(mask_shape)),
    }

    if temp_dir:
        path = os.path.join(temp_dir, "image.png")
        cv2.imwrite(path, np.ones(image_shape))
        return dm.DatasetItem(id=item_id, subset=subset, image=path, annotations=[ann_task_dict[task]])

    return dm.DatasetItem(id=item_id, subset=subset, image=np.ones(image_shape), annotations=[ann_task_dict[task]])


def generate_datumaro_dataset(
    subsets: List[str],
    task: str,
    num_data: int = 1,
    image_shape: np.array = np.array((5, 5, 3)),
    mask_shape: np.array = np.array((5, 5)),
) -> dm.Dataset:
    """Generate Datumaro Dataset.

    Args:
        subsets (List): the list of subset, e.g. ["train", "val"]
        task (str): task name, e.g. "classification", "segmentation", ..
        num_data (int): the number of dataset to make.
        image_shape (np.array): the shape of image.
        mask_shape (np.array): the shape of mask.

    Returns:
        dm.Dataset: Datumaro Dataset
    """
    dataset_items: dm.DatasetItem = []
    for subset in subsets:
        for idx in range(num_data):
            dataset_items.append(
                generate_datumaro_dataset_item(
                    item_id=f"{subset}/image{idx}",
                    subset=subset,
                    task=task,
                    image_shape=image_shape,
                    mask_shape=mask_shape,
                )
            )
    return dm.Dataset.from_iterable(dataset_items, categories=["cat", "dog"])



"""OTX Adapters - mmdet.utils."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from .builder import build_classifier
from .config_utils import patch_datasets, patch_evaluation

__all__ = [
    "build_classifier",
    "patch_datasets",
    "patch_evaluation",
]



"""MMDeploy config partitioning EfficientNetB2B MaskRCNN model to tile classifier and MaskRCNN model."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

_base_ = ["./deployment.py"]

ir_config = dict(
    output_names=["boxes", "labels", "masks", "tile_prob"],
)

partition_config = dict(
    type="tile_classifier",
    apply_marks=True,
    partition_cfg=[
        dict(
            save_file="tile_classifier.onnx",
            start=["tile_classifier:input"],
            end=["tile_classifier:output"],
            output_names=["tile_prob"],
        )
    ],
)



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import datetime

import numpy as np
import pytest

from otx.api.configuration import ConfigurableParameters
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.color import Color
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelGroup, LabelSchemaEntity
from otx.api.entities.metrics import (
    BarChartInfo,
    BarMetricsGroup,
    ColorPalette,
    Performance,
    ScoreMetric,
    VisualizationType,
)
from otx.api.entities.model import ModelConfiguration, ModelEntity
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.usecases.evaluation.averaging import MetricAverageMethod
from otx.api.usecases.evaluation.dice import DiceAverage
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestDice:
    color = Color(0, 255, 0)
    creation_date = datetime.datetime(year=2022, month=1, day=10)
    image = Image(data=np.random.randint(low=0, high=255, size=(64, 32, 3)))
    full_box_roi = Rectangle.generate_full_box()
    car_label = LabelEntity(
        name="car",
        domain=Domain.DETECTION,
        color=color,
        creation_date=creation_date,
        id=ID("car_label"),
    )
    human_label = LabelEntity(
        name="human",
        domain=Domain.DETECTION,
        color=color,
        creation_date=creation_date,
        id=ID("human_label"),
    )
    dog_label = LabelEntity(
        name="dog",
        domain=Domain.DETECTION,
        color=color,
        creation_date=creation_date,
        id=ID("dog_label"),
    )
    cat_label = LabelEntity(
        name="cat",
        domain=Domain.DETECTION,
        color=color,
        creation_date=creation_date,
        id=ID("cat_label"),
    )
    configurable_params = ConfigurableParameters(header="Test model configurable params")

    def human_1_ground_truth(self) -> DatasetItemEntity:
        human_roi = Annotation(shape=self.full_box_roi, labels=[ScoredLabel(self.human_label)])
        human_annotation = Annotation(
            shape=Rectangle(x1=0.4, y1=0, x2=0.5, y2=0.2),
            labels=[ScoredLabel(self.human_label)],
        )
        human_annotation_scene = AnnotationSceneEntity(
            annotations=[human_annotation], kind=AnnotationSceneKind.ANNOTATION
        )
        human_dataset_item = DatasetItemEntity(media=self.image, annotation_scene=human_annotation_scene, roi=human_roi)
        return human_dataset_item

    def human_2_ground_truth(self) -> DatasetItemEntity:
        human_roi = Annotation(shape=self.full_box_roi, labels=[ScoredLabel(self.human_label)])
        human_annotation = Annotation(
            shape=Rectangle(x1=0.6, y1=0, x2=0.7, y2=0.2),
            labels=[ScoredLabel(self.human_label)],
        )
        human_annotation_scene = AnnotationSceneEntity(
            annotations=[human_annotation], kind=AnnotationSceneKind.ANNOTATION
        )
        human_dataset_item = DatasetItemEntity(media=self.image, annotation_scene=human_annotation_scene, roi=human_roi)
        return human_dataset_item

    def dog_ground_truth(self) -> DatasetItemEntity:
        dog_roi = Annotation(shape=self.full_box_roi, labels=[ScoredLabel(self.dog_label)])
        dog_annotation = Annotation(
            shape=Rectangle(x1=0.8, y1=0, x2=0.9, y2=0.1),
            labels=[ScoredLabel(self.dog_label)],
        )
        dog_annotation_scene = AnnotationSceneEntity(annotations=[dog_annotation], kind=AnnotationSceneKind.ANNOTATION)
        dog_dataset_item = DatasetItemEntity(media=self.image, annotation_scene=dog_annotation_scene, roi=dog_roi)
        return dog_dataset_item

    def human_1_predicted(self) -> DatasetItemEntity:
        human_roi = Annotation(shape=self.full_box_roi, labels=[ScoredLabel(self.human_label)])
        human_annotation = Annotation(
            shape=Rectangle(x1=0.4, y1=0, x2=0.5, y2=0.4),
            labels=[ScoredLabel(self.human_label)],
        )
        human_annotation_scene = AnnotationSceneEntity(
            annotations=[human_annotation], kind=AnnotationSceneKind.ANNOTATION
        )
        human_dataset_item = DatasetItemEntity(media=self.image, annotation_scene=human_annotation_scene, roi=human_roi)
        return human_dataset_item

    def human_2_predicted(self) -> DatasetItemEntity:
        human_roi = Annotation(shape=self.full_box_roi, labels=[ScoredLabel(self.human_label)])
        human_annotation = Annotation(
            shape=Rectangle(x1=0.6, y1=0, x2=0.7, y2=0.2),
            labels=[ScoredLabel(self.human_label)],
        )
        human_annotation_scene = AnnotationSceneEntity(
            annotations=[human_annotation], kind=AnnotationSceneKind.ANNOTATION
        )
        human_dataset_item = DatasetItemEntity(media=self.image, annotation_scene=human_annotation_scene, roi=human_roi)
        return human_dataset_item

    def cat_predicted(self) -> DatasetItemEntity:
        cat_roi = Annotation(shape=self.full_box_roi, labels=[ScoredLabel(self.cat_label)])
        cat_annotation = Annotation(
            shape=Rectangle(x1=0.9, y1=0, x2=1.0, y2=0.1),
            labels=[ScoredLabel(self.cat_label)],
        )
        cat_annotation_scene = AnnotationSceneEntity(annotations=[cat_annotation], kind=AnnotationSceneKind.ANNOTATION)
        cat_dataset_item = DatasetItemEntity(media=self.image, annotation_scene=cat_annotation_scene, roi=cat_roi)
        return cat_dataset_item

    def car_dataset_item(self) -> DatasetItemEntity:
        car_roi = Annotation(
            shape=self.full_box_roi,
            labels=[ScoredLabel(self.car_label)],
        )
        car_annotation = Annotation(
            shape=Rectangle(x1=0.1, y1=0, x2=0.3, y2=0.2),
            labels=[ScoredLabel(self.car_label)],
        )
        car_annotation_scene = AnnotationSceneEntity(annotations=[car_annotation], kind=AnnotationSceneKind.ANNOTATION)
        car_dataset_item = DatasetItemEntity(media=self.image, annotation_scene=car_annotation_scene, roi=car_roi)
        return car_dataset_item

    def model(self) -> ModelEntity:
        labels_group = LabelGroup(
            name="model_labels_group",
            labels=[self.car_label, self.human_label, self.dog_label, self.cat_label],
        )
        model_configuration = ModelConfiguration(
            configurable_parameters=self.configurable_params,
            label_schema=LabelSchemaEntity(label_groups=[labels_group]),
        )
        model = ModelEntity(train_dataset=DatasetEntity(), configuration=model_configuration)
        return model

    @staticmethod
    def check_score_metric(score_metric, expected_name, expected_value):
        assert isinstance(score_metric, ScoreMetric)
        assert score_metric.name == expected_name
        assert score_metric.value == pytest.approx(expected_value)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dice_initialization(self):
        """
        <b>Description:</b>
        Check "DiceAverage" class object initialization

        <b>Input data:</b>
        "DiceAverage" object with specified "resultset" and "average" parameters

        <b>Expected results:</b>
        Test passes if attributes of initialized "DiceAverage" object are equal to expected

        <b>Steps</b>
        1. Check attributes of "DiceAverage" object initialized with default value of "average" parameter
        2. Check attributes of "DiceAverage" object initialized with specified value of "average" parameter
        3. Check that "ValueError" exception is raised when initializing "DiceAverage" object with empty list prediction
        "resultset" attribute
        4. Check "ValueError" exception is raised when initializing "DiceAverage" object with "resultset" attribute with
        unequal length of "ground_truth_dataset" and "prediction_dataset"
        """

        def check_dice_attributes(
            dice_actual: DiceAverage,
            expected_average_type: MetricAverageMethod,
            expected_overall_dice: float,
        ):
            assert dice_actual.average == expected_average_type
            # Checking "overall_dice" attribute
            self.check_score_metric(
                score_metric=dice_actual.overall_dice,
                expected_name="Dice Average",
                expected_value=expected_overall_dice,
            )
            # Checking "dice_per_label" attribute
            assert len(dice_actual.dice_per_label) == 4
            self.check_score_metric(
                score_metric=dice_actual.dice_per_label.get(self.car_label),
                expected_name="car",
                expected_value=1.0,
            )
            self.check_score_metric(
                score_metric=dice_actual.dice_per_label.get(self.human_label),
                expected_name="human",
                expected_value=0.782608695652174,
            )
            self.check_score_metric(
                score_metric=dice_actual.dice_per_label.get(self.dog_label),
                expected_name="dog",
                expected_value=0.0,
            )
            self.check_score_metric(
                score_metric=dice_actual.dice_per_label.get(self.cat_label),
                expected_name="cat",
                expected_value=0.0,
            )

        model = self.model()
        human_1_ground_truth = self.human_1_ground_truth()
        car_dataset_item = self.car_dataset_item()
        ground_truth_dataset = DatasetEntity(
            [
                car_dataset_item,
                human_1_ground_truth,
                self.human_2_ground_truth(),
                self.dog_ground_truth(),
            ]
        )
        predicted_dataset = DatasetEntity(
            [
                car_dataset_item,
                self.human_1_predicted(),
                self.human_2_predicted(),
                self.cat_predicted(),
            ]
        )
        result_set = ResultSetEntity(
            model=model,
            ground_truth_dataset=ground_truth_dataset,
            prediction_dataset=predicted_dataset,
        )
        # Checking attributes of "DiceAverage" initialized with default "average"
        dice = DiceAverage(resultset=result_set)
        check_dice_attributes(
            dice_actual=dice,
            expected_average_type=MetricAverageMethod.MACRO,
            expected_overall_dice=0.44565217391304346,
        )
        # Checking attributes of "DiceAverage" initialized with specified "average"
        dice = DiceAverage(resultset=result_set, average=MetricAverageMethod.MICRO)
        check_dice_attributes(
            dice_actual=dice,
            expected_average_type=MetricAverageMethod.MICRO,
            expected_overall_dice=0.7746741154562383,
        )
        # Checking "ValueError" exception raised when initializing "DiceAverage" with empty list prediction result_set
        result_set = ResultSetEntity(
            model=model,
            ground_truth_dataset=ground_truth_dataset,
            prediction_dataset=DatasetEntity(items=[]),
        )
        with pytest.raises(ValueError):
            DiceAverage(resultset=result_set)

        # Checking "ValueError" exception raised when initializing "DiceAverage" with "resultset" with unequal length of
        # "ground_truth_dataset" and "prediction_dataset"
        result_set = ResultSetEntity(
            model=model,
            ground_truth_dataset=DatasetEntity([car_dataset_item, human_1_ground_truth]),
            prediction_dataset=DatasetEntity([car_dataset_item]),
        )
        with pytest.raises(ValueError):
            DiceAverage(resultset=result_set)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dice_get_performance(self):
        """
        <b>Description:</b>
        Check "DiceAverage" class "get_performance" method

        <b>Input data:</b>
        "DiceAverage" object with specified "resultset" and "average" parameters

        <b>Expected results:</b>
        Test passes if "Performance" object returned by "get_performance" method is equal to expected

        <b>Steps</b>
        1. Check "Performance" object returned by "get_performance" method for "DiceAverage" class object with length of
        "dice_per_label" attribute more than 0
        2. Check "Performance" object returned by "get_performance" method for "DiceAverage" class object with length of
        "dice_per_label" attribute equal to 0
        """
        human_1_ground_truth = self.human_1_ground_truth()
        human_1_predicted = self.human_1_predicted()
        car_dataset_item = self.car_dataset_item()
        # Checking "Performance" returned by "get_performance" for "DiceAverage" with length of "dice_per_label" more
        # than 0
        configurable_params = self.configurable_params
        labels_group = LabelGroup(
            name="model_labels_group",
            labels=[self.car_label, self.human_label, self.dog_label, self.cat_label],
        )
        model_configuration = ModelConfiguration(
            configurable_parameters=configurable_params,
            label_schema=LabelSchemaEntity(label_groups=[labels_group]),
        )
        model = ModelEntity(train_dataset=DatasetEntity(), configuration=model_configuration)
        ground_truth_dataset = DatasetEntity(
            [
                car_dataset_item,
                human_1_ground_truth,
                self.human_2_ground_truth(),
                self.dog_ground_truth(),
            ]
        )
        predicted_dataset = DatasetEntity(
            [
                car_dataset_item,
                human_1_predicted,
                self.human_2_predicted(),
                self.cat_predicted(),
            ]
        )
        result_set = ResultSetEntity(
            model=model,
            ground_truth_dataset=ground_truth_dataset,
            prediction_dataset=predicted_dataset,
        )
        dice = DiceAverage(resultset=result_set)
        performance = dice.get_performance()
        assert isinstance(performance, Performance)
        # Checking "score" attribute
        self.check_score_metric(
            score_metric=performance.score,
            expected_name="Dice Average",
            expected_value=0.44565217391304346,
        )
        # Checking "dashboard_metrics" attribute
        assert len(performance.dashboard_metrics) == 1
        dashboard_metric = performance.dashboard_metrics[0]
        assert isinstance(dashboard_metric, BarMetricsGroup)
        # Checking "metrics" attribute
        assert len(dashboard_metric.metrics) == 4
        self.check_score_metric(
            score_metric=dashboard_metric.metrics[0],
            expected_name="car",
            expected_value=1.0,
        )
        self.check_score_metric(
            score_metric=dashboard_metric.metrics[1],
            expected_name="cat",
            expected_value=0.0,
        )
        self.check_score_metric(
            score_metric=dashboard_metric.metrics[2],
            expected_name="dog",
            expected_value=0.0,
        )
        self.check_score_metric(
            score_metric=dashboard_metric.metrics[3],
            expected_name="human",
            expected_value=0.782608695652174,
        )
        # Checking "visualization_info" attribute
        assert isinstance(dashboard_metric.visualization_info, BarChartInfo)
        assert dashboard_metric.visualization_info.name == "Dice Average Per Label"
        assert dashboard_metric.visualization_info.palette == ColorPalette.LABEL
        assert dashboard_metric.visualization_info.type == VisualizationType.BAR
        # Checking "Performance" returned by "get_performance" for "DiceAverage" with length of "dice_per_label" equal
        # to 0
        labels_group = LabelGroup(name="model_labels_group", labels=[self.car_label])
        model_configuration = ModelConfiguration(configurable_params, LabelSchemaEntity(label_groups=[labels_group]))
        model = ModelEntity(train_dataset=DatasetEntity(), configuration=model_configuration)
        ground_truth_dataset = DatasetEntity([human_1_ground_truth])
        predicted_dataset = DatasetEntity([human_1_predicted])
        result_set = ResultSetEntity(
            model=model,
            ground_truth_dataset=ground_truth_dataset,
            prediction_dataset=predicted_dataset,
        )
        dice = DiceAverage(resultset=result_set)
        performance = dice.get_performance()
        assert isinstance(performance, Performance)
        # Checking "score" attribute
        self.check_score_metric(
            score_metric=performance.score,
            expected_name="Dice Average",
            expected_value=0.0,
        )
        # Checking "dashboard_metrics" attribute
        assert performance.dashboard_metrics == []

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dice_compute_dice_using_intersection_and_cardinality(self):
        """
        <b>Description:</b>
        Check "DiceAverage" class "compute_dice_using_intersection_and_cardinality" method

        <b>Input data:</b>
        "DiceAverage" object, "all_intersection" dictionary, "all_cardinality" dictionary, "average" MetricAverageMethod
        object

        <b>Expected results:</b>
        Test passes if tuple returned by "compute_dice_using_intersection_and_cardinality" method is equal to expected

        <b>Steps</b>
        1. Check tuple returned by "compute_dice_using_intersection_and_cardinality" method for MICRO "average"
        parameter
        2. Check tuple returned by "compute_dice_using_intersection_and_cardinality" method for MACRO "average"
        parameter
        3. Check tuple returned by "compute_dice_using_intersection_and_cardinality" method when "None" key is not
        specified in "all_intersection" and "all_cardinality" dictionaries
        4. Check "KeyError" exception is raised when keys of "all_intersection" and "all_cardinality" dictionaries are
        not match
        5. Check "ValueError" exception is raised when intersection of certain key is larger than its cardinality
        """

        def check_dice(dice_actual: tuple, expected_overall_dice: float):
            assert len(dice_actual) == 2
            self.check_score_metric(
                score_metric=dice_actual[0],
                expected_name="Dice Average",
                expected_value=expected_overall_dice,
            )
            dice_per_label = dice_actual[1]
            assert len(dice_per_label) == 4
            self.check_score_metric(
                score_metric=dice_per_label.get(self.car_label),
                expected_name="car",
                expected_value=2.0,
            )
            self.check_score_metric(
                score_metric=dice_per_label.get(self.human_label),
                expected_name="human",
                expected_value=1.6,
            )
            self.check_score_metric(
                score_metric=dice_per_label.get(self.dog_label),
                expected_name="dog",
                expected_value=0.5,
            )
            self.check_score_metric(
                score_metric=dice_per_label.get(self.cat_label),
                expected_name="cat",
                expected_value=0.0,
            )

        # Checking tuple returned by "compute_dice_using_intersection_and_cardinality" for MICRO "average"
        all_intersection = {
            self.car_label: 10,
            self.human_label: 8,
            self.dog_label: 2,
            self.cat_label: 0,
            None: 9,
        }
        all_cardinality = {
            self.car_label: 10,
            self.human_label: 10,
            self.dog_label: 8,
            self.cat_label: 2,
            None: 12,
        }
        dice = DiceAverage.compute_dice_using_intersection_and_cardinality(
            all_intersection=all_intersection,
            all_cardinality=all_cardinality,
            average=MetricAverageMethod.MICRO,
        )
        check_dice(dice_actual=dice, expected_overall_dice=1.5)
        # Checking tuple returned by "compute_dice_using_intersection_and_cardinality" for MACRO "average"
        dice = DiceAverage.compute_dice_using_intersection_and_cardinality(
            all_intersection=all_intersection,
            all_cardinality=all_cardinality,
            average=MetricAverageMethod.MACRO,
        )
        check_dice(dice_actual=dice, expected_overall_dice=1.025)
        # Checking tuple returned by "compute_dice_using_intersection_and_cardinality" when "None" key is not
        # specified in "all_intersection" and "all_cardinality" dictionaries
        all_intersection = {
            self.car_label: 10,
            self.human_label: 8,
            self.dog_label: 2,
            self.cat_label: 0,
        }
        all_cardinality = {
            self.car_label: 10,
            self.human_label: 10,
            self.dog_label: 8,
            self.cat_label: 2,
        }
        # Check for MACRO "average" parameter
        dice = DiceAverage.compute_dice_using_intersection_and_cardinality(
            all_intersection=all_intersection,
            all_cardinality=all_cardinality,
            average=MetricAverageMethod.MACRO,
        )
        check_dice(dice_actual=dice, expected_overall_dice=1.025)
        # Expected KeyError exception for MICRO "average" parameter
        with pytest.raises(KeyError):
            DiceAverage.compute_dice_using_intersection_and_cardinality(
                all_intersection=all_intersection,
                all_cardinality=all_cardinality,
                average=MetricAverageMethod.MICRO,
            )
        # Checking "KeyError" exception is raised when keys of "all_intersection" and "all_cardinality" dictionaries are
        # not match
        all_intersection = {self.car_label: 10, self.human_label: 9, None: 9}
        all_cardinality = {
            self.car_label: 10,
            self.dog_label: 8,
            self.cat_label: 2,
            None: 12,
        }
        with pytest.raises(KeyError):
            DiceAverage.compute_dice_using_intersection_and_cardinality(
                all_intersection=all_intersection,
                all_cardinality=all_cardinality,
                average=MetricAverageMethod.MACRO,
            )
        # Checking "ValueError" exception is raised when intersection of certain key is larger than its cardinality
        all_intersection = {self.car_label: 10, self.human_label: 9, None: 12}
        all_cardinality = {self.car_label: 10, self.human_label: 8, None: 12}
        with pytest.raises(ValueError):
            DiceAverage.compute_dice_using_intersection_and_cardinality(
                all_intersection=all_intersection,
                all_cardinality=all_cardinality,
                average=MetricAverageMethod.MACRO,
            )



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.algorithms.detection.adapters.mmdet.models.roi_heads.roi_extractors.single_level_roi_extractor import (
    RoIInterpolationPool,
    SingleRoIExtractor,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestSingleRoIExtractor:
    @e2e_pytest_unit
    def test_build_roi_layers(self):
        extractor = SingleRoIExtractor(
            roi_layer=dict(type="RoIInterpolationPool", output_size=14, mode="bilinear"),
            out_channels=1024,
            featmap_strides=[8],
        )
        assert all(isinstance(layer, RoIInterpolationPool) for layer in extractor.roi_layers)



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from time import time

import pytest

from otx.api.entities.train_parameters import default_progress_callback
from otx.api.usecases.reporting.time_monitor_callback import TimeMonitorCallback
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestTimeMonitorCallback:
    @staticmethod
    def time_monitor_callback():
        return TimeMonitorCallback(
            num_epoch=3,
            num_train_steps=4,
            num_val_steps=5,
            num_test_steps=6,
            update_progress_callback=default_progress_callback,
        )

    @staticmethod
    def check_current_step_start_step_time_attributes(
        callback: TimeMonitorCallback,
        expected_step: int,
        expected_step_time_before: float,
    ):
        assert callback.current_step == expected_step
        assert callback.start_step_time > expected_step_time_before

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_time_monitor_callback_initialization(self):
        """
        <b>Description:</b>
        Check "TimeMonitorCallback" class object initialization

        <b>Input data:</b>
        "TimeMonitorCallback" class object with specified initialization parameters

        <b>Expected results:</b>
        Test passes if attributes of initialized "TimeMonitorCallback" class object are equal to expected

        <b>Steps</b>
        1. Check attributes of "TimeMonitorCallback" class object initialized with default optional parameters
        2. Check attributes of "TimeMonitorCallback" class object initialized with specified optional parameters
        """

        def check_time_monitor_callback_attributes(
            actual_time_monitor_callback: TimeMonitorCallback,
            expected_step_history: int,
            expected_epoch_history: int,
            expected_update_progress_callback,
        ):
            assert actual_time_monitor_callback.total_epochs == 3
            assert actual_time_monitor_callback.train_steps == 4
            assert actual_time_monitor_callback.val_steps == 5
            assert actual_time_monitor_callback.test_steps == 6
            assert actual_time_monitor_callback.steps_per_epoch == 9  # train_steps + val_steps
            assert actual_time_monitor_callback.total_steps == 33  # steps_per_epoch * total_epochs + num_test_steps
            assert actual_time_monitor_callback.current_step == 0
            assert actual_time_monitor_callback.current_epoch == 0
            assert actual_time_monitor_callback.start_step_time  # calculated using time.time() method
            assert actual_time_monitor_callback.past_step_duration == []
            assert actual_time_monitor_callback.average_step == 0
            assert actual_time_monitor_callback.step_history == expected_step_history
            assert actual_time_monitor_callback.start_epoch_time  # calculated using time.time() method
            assert actual_time_monitor_callback.past_epoch_duration == []
            assert actual_time_monitor_callback.average_epoch == 0
            assert actual_time_monitor_callback.epoch_history == expected_epoch_history
            assert not actual_time_monitor_callback.is_training
            assert actual_time_monitor_callback.update_progress_callback == expected_update_progress_callback

        # Checking attributes of "TimeMonitorCallback" initialized with default optional parameters
        num_epoch = 3
        num_train_steps = 4
        num_val_steps = 5
        num_test_steps = 6

        time_monitor_callback = TimeMonitorCallback(
            num_epoch=num_epoch,
            num_train_steps=num_train_steps,
            num_val_steps=num_val_steps,
            num_test_steps=num_test_steps,
        )
        check_time_monitor_callback_attributes(
            actual_time_monitor_callback=time_monitor_callback,
            expected_epoch_history=5,
            expected_step_history=50,
            expected_update_progress_callback=default_progress_callback,
        )
        # Checking attributes of "TimeMonitorCallback" initialized with specified optional parameters
        step_history = 10
        epoch_history = 100
        callback = default_progress_callback
        time_monitor_callback = TimeMonitorCallback(
            num_epoch=num_epoch,
            num_train_steps=num_train_steps,
            num_val_steps=num_val_steps,
            num_test_steps=num_test_steps,
            step_history=step_history,
            epoch_history=epoch_history,
            update_progress_callback=callback,
        )
        check_time_monitor_callback_attributes(
            actual_time_monitor_callback=time_monitor_callback,
            expected_step_history=step_history,
            expected_epoch_history=epoch_history,
            expected_update_progress_callback=callback,
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_time_monitor_callback_on_train_batch_begin(self):
        """
        <b>Description:</b>
        Check "TimeMonitorCallback" class object "on_train_batch_begin" method

        <b>Input data:</b>
        "TimeMonitorCallback" class object with specified initialization parameters

        <b>Expected results:</b>
        Test passes if "current_step" and "start_step_time" attributes of "TimeMonitorCallback" class object after
        "on_train_batch_begin" method are equal to expected
        """

        time_monitor_callback = self.time_monitor_callback()
        start_step_time_before = time() - 1
        time_monitor_callback.start_step_time = start_step_time_before
        time_monitor_callback.on_train_batch_begin(batch=1, logs="on_train_batch_begin logs")
        self.check_current_step_start_step_time_attributes(
            callback=time_monitor_callback,
            expected_step=1,
            expected_step_time_before=start_step_time_before,
        )
        time_monitor_callback.start_step_time = start_step_time_before
        time_monitor_callback.on_train_batch_begin(batch=2, logs="on_train_batch_begin logs")
        self.check_current_step_start_step_time_attributes(
            callback=time_monitor_callback,
            expected_step=2,
            expected_step_time_before=start_step_time_before,
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_time_monitor_callback_on_train_batch_end(self):
        """
        <b>Description:</b>
        Check "TimeMonitorCallback" class object "on_train_batch_end" method

        <b>Input data:</b>
        "TimeMonitorCallback" class object with specified initialization parameters

        <b>Expected results:</b>
        Test passes if "past_step_duration" and "average_step" attributes of "TimeMonitorCallback" class object after
        "on_train_batch_end" method are equal to expected

        <b>Steps</b>
        1. Check "past_step_duration" and "average_step" attributes after "on_train_batch_end" method for
        "TimeMonitorCallback" class object with length of "past_step_duration" attribute more than value of
        "step_history" attribute
        2. Check "past_step_duration" and "average_step" attributes after "on_train_batch_end" method for
        "TimeMonitorCallback" class object with length of "past_step_duration" attribute less than value of
        "step_history" attribute
        """
        # Checking "past_step_duration" and "average_step" after "on_train_batch_end" for "TimeMonitorCallback" with
        # length of "past_step_duration" more than "step_history" value
        time_monitor_callback = self.time_monitor_callback()
        time_monitor_callback.past_step_duration = [10.0, 12.0]
        time_monitor_callback.step_history = 2
        time_monitor_callback.on_train_batch_end(batch=1, logs="on_train_batch_end logs")
        assert len(time_monitor_callback.past_step_duration) == 2
        assert round(time_monitor_callback.average_step, 4) == 6.0
        # Checking "past_step_duration" and "average_step" after "on_train_batch_end" for "TimeMonitorCallback" with
        # length of "past_step_duration" less than "step_history" value
        time_monitor_callback = self.time_monitor_callback()
        time_monitor_callback.past_step_duration = [10, 14]
        time_monitor_callback.step_history = 4
        time_monitor_callback.on_train_batch_end(batch=2, logs="on_train_batch_end logs")
        assert len(time_monitor_callback.past_step_duration) == 3
        assert round(time_monitor_callback.average_step, 4) == 8.0

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_time_monitor_callback_is_stalling(self):
        """
        <b>Description:</b>
        Check "TimeMonitorCallback" class object "is_stalling" method

        <b>Input data:</b>
        "TimeMonitorCallback" class object with specified initialization parameters

        <b>Expected results:</b>
        Test passes if bool value returned by "is_stalling" method is equal to expected

        <b>Steps</b>
        1. Check value returned by "is_stalling" method for "TimeMonitorCallback" object with "is_training" attribute
        is "True" and "current_step" more than 2
        2. Check value returned by "is_stalling" method for "TimeMonitorCallback" object with "is_training" attribute
        is "True" and "current_step" equal to 2
        3. Check value returned by "is_stalling" method for "TimeMonitorCallback" object with "is_training" attribute
        is "False" and "current_step" more than 2
        """
        time_monitor_callback = self.time_monitor_callback()
        # Checking value returned by "is_stalling" for "TimeMonitorCallback" with "is_training" is "True" and
        # "current_step" more than 2
        time_monitor_callback.is_training = True
        time_monitor_callback.current_step = 3
        time_monitor_callback.start_step_time = 40
        assert time_monitor_callback.is_stalling()
        # Checking value returned by "is_stalling" for "TimeMonitorCallback" with "is_training" is "True" and
        # "current_step" equal to 2
        time_monitor_callback.current_step = 2
        assert not time_monitor_callback.is_stalling()
        # Checking value returned by "is_stalling" for "TimeMonitorCallback" with "is_training" is "False" and
        # "current_step" more than 2
        time_monitor_callback.current_step = 3
        time_monitor_callback.is_training = False
        assert not time_monitor_callback.is_stalling()

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_time_monitor_callback_on_test_batch_begin(self):
        """
        <b>Description:</b>
        Check "TimeMonitorCallback" class object "on_test_batch_begin" method

        <b>Input data:</b>
        "TimeMonitorCallback" class object with specified initialization parameters

        <b>Expected results:</b>
        Test passes if "current_step" and "start_step_time" attributes of "TimeMonitorCallback" class object after
        "on_test_batch_begin" method are equal to expected
        """
        time_monitor_callback = self.time_monitor_callback()
        start_step_time_before = time() - 1
        time_monitor_callback.start_step_time = start_step_time_before
        time_monitor_callback.on_test_batch_begin(batch=1, logs="on_test_batch_begin logs")
        self.check_current_step_start_step_time_attributes(time_monitor_callback, 1, start_step_time_before)
        time_monitor_callback.start_step_time = start_step_time_before
        time_monitor_callback.on_test_batch_begin(batch=2, logs="on_test_batch_begin logs")
        self.check_current_step_start_step_time_attributes(time_monitor_callback, 2, start_step_time_before)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_time_monitor_callback_on_test_batch_end(self):
        """
        <b>Description:</b>
        Check "TimeMonitorCallback" class object "on_test_batch_end" method

        <b>Input data:</b>
        "TimeMonitorCallback" class object with specified initialization parameters

        <b>Expected results:</b>
        Test passes if "current_step" and "start_step_time" attributes of "TimeMonitorCallback" class object after
        "on_test_batch_end" method are equal to expected

        <b>Steps</b>
        1. Check "past_step_duration" and "average_step" attributes after "on_test_batch_end" method for
        "TimeMonitorCallback" class object with length of "past_step_duration" attribute more than value of
        "step_history" attribute
        2. Check "past_step_duration" and "average_step" attributes after "on_test_batch_end" method for
        "TimeMonitorCallback" class object with length of "past_step_duration" attribute less than value of
        "step_history" attribute
        """
        # Checking "past_step_duration" and "average_step" after "on_train_batch_end" for "TimeMonitorCallback" with
        # length of "past_step_duration" more than "step_history" value
        time_monitor_callback = self.time_monitor_callback()
        time_monitor_callback.past_step_duration = [5.0, 6.0]
        time_monitor_callback.step_history = 2
        time_monitor_callback.on_test_batch_end(batch=1, logs="on_test_batch_end logs")
        assert len(time_monitor_callback.past_step_duration) == 2
        assert round(time_monitor_callback.average_step, 4) == 3.0
        # Checking "past_step_duration" and "average_step" after "on_train_batch_end" for "TimeMonitorCallback" with
        # length of "past_step_duration" less than "step_history" value
        time_monitor_callback = self.time_monitor_callback()
        time_monitor_callback.past_step_duration = [10.0, 14.0, 16.0]
        time_monitor_callback.step_history = 5
        time_monitor_callback.on_test_batch_end(batch=2, logs="on_test_batch_end logs")
        assert len(time_monitor_callback.past_step_duration) == 4
        assert round(time_monitor_callback.average_step, 4) == 10.0

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_time_monitor_callback_on_train_begin(self):
        """
        <b>Description:</b>
        Check "TimeMonitorCallback" class object "on_train_begin" method

        <b>Input data:</b>
        "TimeMonitorCallback" class object with specified initialization parameters

        <b>Expected results:</b>
        Test passes if "is_training" attribute of "TimeMonitorCallback" class object after "on_train_begin" method is
        "True"

        <b>Steps</b>
        1. Check "is_training" attribute after "on_train_begin" method for "TimeMonitorCallback" class object with
        "is_training" attribute is "False"
        2. Check "is_training" attribute after "on_train_begin" method for "TimeMonitorCallback" class object with
        "is_training" attribute is "True"
        """
        # Checking "is_training" after "on_train_begin" for "TimeMonitorCallback" with "is_training" is "False"
        time_monitor_callback = self.time_monitor_callback()
        time_monitor_callback.on_train_begin("on_train_begin logs")
        assert time_monitor_callback.is_training
        # Checking "is_training" after "on_train_begin" for "TimeMonitorCallback" with "is_training" is "True"
        time_monitor_callback.on_train_begin("on_train_begin logs")
        assert time_monitor_callback.is_training

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_time_monitor_callback_on_train_end(self):
        """
        <b>Description:</b>
        Check "TimeMonitorCallback" class object "on_train_end" method

        <b>Input data:</b>
        "TimeMonitorCallback" class object with specified initialization parameters

        <b>Expected results:</b>
        Test passes if "current_step", "current_epoch" and "is_training" attributes of "TimeMonitorCallback" class
        object after "on_train_end" method are equal to expected

        <b>Steps</b>
        1. Check "current_step", "current_epoch" and "is_training" attributes after "on_train_end" method for
        "TimeMonitorCallback" class object with "is_training" attribute is "True"
        2. Check "is_training" attribute after "on_train_begin" method for "TimeMonitorCallback" class object with
        "is_training" attribute is "False"
        """

        def check_attributes_after_on_train_end(callback: TimeMonitorCallback):
            assert callback.current_step == 27  # total_steps - test_steps

            assert callback.current_epoch == 3  # total_epochs
            assert not callback.is_training

        # Checking "current_step", "current_epoch" and "is_training" after "on_train_end" for "TimeMonitorCallback" with
        # "is_training": "True"
        time_monitor_callback = self.time_monitor_callback()
        time_monitor_callback.on_train_begin("on_train_begin logs")  # setting "is_training" to "True"
        time_monitor_callback.on_train_end("on_train_end logs")
        check_attributes_after_on_train_end(time_monitor_callback)
        # Checking "current_step", "current_epoch" and "is_training" after "on_train_end" for "TimeMonitorCallback" with
        # "is_training": "False"
        time_monitor_callback = self.time_monitor_callback()
        time_monitor_callback.on_train_end("on_train_end logs")
        check_attributes_after_on_train_end(time_monitor_callback)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_time_monitor_callback_on_epoch_begin(self):
        """
        <b>Description:</b>
        Check "TimeMonitorCallback" class object "on_epoch_begin" method

        <b>Input data:</b>
        "TimeMonitorCallback" class object with specified initialization parameters

        <b>Expected results:</b>
        Test passes if "current_epoch" and "start_epoch_time" attributes of "TimeMonitorCallback" class object after
        "on_epoch_begin" method are equal to expected
        """
        start_epoch_time = time() - 1
        time_monitor_callback = self.time_monitor_callback()
        time_monitor_callback.start_epoch_time = start_epoch_time
        time_monitor_callback.on_epoch_begin(epoch=0, logs="on_epoch_begin logs")
        assert time_monitor_callback.current_epoch == 1
        assert time_monitor_callback.start_epoch_time > start_epoch_time
        time_monitor_callback.start_epoch_time = start_epoch_time
        time_monitor_callback.on_epoch_begin(epoch=2, logs="on_epoch_begin logs")
        assert time_monitor_callback.current_epoch == 3
        assert time_monitor_callback.start_epoch_time > start_epoch_time

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_time_monitor_callback_on_epoch_end(self):
        """
        <b>Description:</b>
        Check "TimeMonitorCallback" class object "on_epoch_end" method

        <b>Input data:</b>
        "TimeMonitorCallback" class object with specified initialization parameters

        <b>Expected results:</b>
        Test passes if "past_epoch_duration" and "average_epoch" attributes of "TimeMonitorCallback" class object after
        "on_epoch_end" method are equal to expected

        <b>Steps</b>
        1. Check "past_epoch_duration" and "average_epoch" attributes after "on_test_batch_end" method for
        "TimeMonitorCallback" class object with length of "past_epoch_duration" attribute more than value of
        "epoch_history" attribute
        2. Check "past_epoch_duration" and "average_epoch" attributes after "on_test_batch_end" method for
        "TimeMonitorCallback" class object with length of "past_epoch_duration" attribute less than value of
        "epoch_history" attribute
        """
        # Checking "past_epoch_duration" and "average_epoch" after "on_test_batch_end" for "TimeMonitorCallback" with
        # length of "past_epoch_duration" more than "epoch_history" value
        time_monitor_callback = self.time_monitor_callback()
        time_monitor_callback.past_epoch_duration = [4.0, 5.0, 10.0]
        time_monitor_callback.epoch_history = 3
        time_monitor_callback.on_epoch_end(epoch=1, logs="on_epoch_end logs")
        assert len(time_monitor_callback.past_epoch_duration) == 3
        assert round(time_monitor_callback.average_epoch, 4) == 5.0
        # Checking "past_epoch_duration" and "average_epoch" after "on_test_batch_end" for "TimeMonitorCallback" with
        # length of "past_epoch_duration" less than "epoch_history" value
        time_monitor_callback = self.time_monitor_callback()
        time_monitor_callback.past_epoch_duration = [4.0, 5.0]
        time_monitor_callback.epoch_history = 4
        time_monitor_callback.on_epoch_end(epoch=2, logs="on_epoch_end logs")
        assert len(time_monitor_callback.past_epoch_duration) == 3
        assert round(time_monitor_callback.average_epoch, 4) == 3.0

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_time_monitor_callback_get_progress(self):
        """
        <b>Description:</b>
        Check "TimeMonitorCallback" class object "get_progress" method

        <b>Input data:</b>
        "TimeMonitorCallback" class object with specified initialization parameters

        <b>Expected results:</b>
        Test passes if value returned by "get_progress" method is equal to expected
        """
        time_monitor_callback = self.time_monitor_callback()
        # Checking value returned by "get_progress" for "TimeMonitorCallback" with "current_step" equal to 0
        assert time_monitor_callback.get_progress() == 0.0
        # Checking value returned by "get_progress" for "TimeMonitorCallback" with "current_step" not equal to 0
        time_monitor_callback.current_step = 16
        time_monitor_callback.total_steps = 64
        assert time_monitor_callback.get_progress() == 25.0  # (current_step / total_steps)*100



# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import pytest

from otx.api.entities.color import Color, ColorEntity
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements

red = 40
red_hex = "28"
green = 210
green_hex = "d2"
blue = 43
blue_hex = "2b"
alpha = 255
alpha_hex = "ff"
color_hex = f"{red_hex}{green_hex}{blue_hex}"

color = Color.from_hex_str(color_hex)


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestColor:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_color(self):
        """
        <b>Description:</b>
        Check that Color can correctly return the value

        <b>Expected results:</b>
        Test passes if the results match
        """

        assert color == Color(red=red, green=green, blue=blue, alpha=alpha)
        assert color.hex_str == f"#{color_hex}{alpha_hex}"
        assert type(color.random()) == Color
        assert color.rgb_tuple == (red, green, blue)
        assert color.bgr_tuple == (blue, green, red)
        assert color != ColorEntity
        assert repr(color) == f"Color(red={red}, green={green}, blue={blue}, alpha={alpha})"
        assert color.red == red
        assert color.green == green
        assert color.blue == blue

        color.red = 68
        color.green = 54
        color.blue = 32
        color.alpha = 0
        assert color.hex_str == "#44362000"


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestColorEntity:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_color_entity(self):
        """
        <b>Description:</b>
        Check that ColorEntity raises some exceptions

        <b>Expected results:</b>
        Test passes if the NotImplementedError is raised
        """

        color_entity = ColorEntity

        with pytest.raises(NotImplementedError):
            color_entity.hex_str.__get__(property)

        with pytest.raises(NotImplementedError):
            color_entity.random()

        with pytest.raises(NotImplementedError):
            color_entity.from_hex_str(color_hex)



import os.path as osp
from copy import deepcopy
from typing import Any, Dict

import pytest
import torch
import datumaro as dm

from otx.algorithms.classification.adapters.mmcls.models.classifiers.sam_classifier import (
    ImageClassifier,
    SAMImageClassifier,
)
from otx.api.entities.datasets import DatasetEntity
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class OTXMobileNetV3:
    pass


class OTXEfficientNet:
    pass


class OTXEfficientNetV2:
    pass


class MockModule:
    def __init__(self, name):
        if name == "mobilenet":
            self.backbone = OTXMobileNetV3()
            self._state_dict = {
                "classifier.4.weight": torch.rand(1, 1),
                "classifier.4.bias": torch.rand(1),
                "act.weight": torch.rand(1),
                "someweights": torch.rand(2),
            }
        elif name == "effnetv2":
            self.backbone = OTXEfficientNetV2()
            self._state_dict = {
                "model.classifier.weight": torch.rand(1, 1),
                "model.weight": torch.rand(1),
            }
        elif name == "effnet":
            self.backbone = OTXEfficientNet()
            self._state_dict = {
                "features.weight": torch.rand(1, 1),
                "features.active.weight": torch.rand(1, 1),
                "output.weight": torch.rand(1),
                "output.asl.weight": torch.rand(1),
            }
        self.multilabel = False
        self.hierarchical = False
        self.is_export = False

    def state_dict(self):
        return self._state_dict


class TestSAMImageClassifier:
    @pytest.fixture(autouse=True)
    def setup(self, mocker) -> None:
        mocker.patch.object(ImageClassifier, "__init__", return_value=None)
        SAMImageClassifier._register_state_dict_hook = mocker.MagicMock()
        SAMImageClassifier._register_load_state_dict_pre_hook = mocker.MagicMock()
        self.classifier = SAMImageClassifier()

    @e2e_pytest_unit
    def test_forward_train(self, mocker):
        img = torch.rand(1, 3, 224, 224)
        gt_labels = torch.rand(1, 1)
        self.classifier.extract_feat = mocker.MagicMock()
        self.classifier.head = mocker.MagicMock()
        self.classifier.augments = None
        self.classifier.multilabel = False
        self.classifier.hierarchical = False
        losses = self.classifier.forward_train(img, gt_labels)

        assert losses is not None

    @pytest.mark.parametrize("name", ["mobilenet", "effnetv2", "effnet"])
    @e2e_pytest_unit
    def test_load_state_dict_pre_hook(self, name):
        self.module = MockModule(name)
        state_dict = self.module.state_dict()
        self.classifier.load_state_dict_pre_hook(self.module, state_dict, prefix="")

        for key in state_dict:
            if name == "mobilenet":
                if "classifier" in key:
                    assert ".3." in key
                if "someweights" in key:
                    assert "backbone." in key
                if "act" in key:
                    assert "head." in key
            elif name == "effnetv2":
                assert "classifier" not in key
                if "model" in key:
                    assert "backbone" in key
            else:
                if "features" in key and "active" not in key:
                    assert "backbone" in key
                elif "active" in key:
                    assert key == "features.active.weight"
                else:
                    assert "head" in key or "fc" in key

    @pytest.mark.parametrize("name", ["mobilenet", "effnetv2", "effnet"])
    @e2e_pytest_unit
    def test_state_dict_hook(self, name):
        self.module = MockModule(name)
        state_dict = self.module.state_dict()
        state_dict_copy = deepcopy(state_dict)
        self.classifier.load_state_dict_pre_hook(self.module, state_dict, prefix="")
        # backward state dict
        self.classifier.state_dict_hook(self.module, state_dict, prefix="")

        assert state_dict.keys() == state_dict_copy.keys()

    @pytest.mark.parametrize("name", ["mobilenet", "effnetv2", "effnet"])
    def test_load_state_dict_mixing_hook(self, name):
        self.module = MockModule(name)
        state_dict = self.module.state_dict()
        chkpt_dict = deepcopy(state_dict)
        model_classes = [0, 1, 2]
        chkpt_classes = [0, 1]
        self.classifier.load_state_dict_mixing_hook(self.module, model_classes, chkpt_classes, chkpt_dict, prefix="")

        assert chkpt_dict.keys() == state_dict.keys()


class TestLossDynamicsTrackingMixin:
    TESTCASE = [
        ("CustomLinearClsHead", "CrossEntropyLoss"),
        ("CustomNonLinearClsHead", "CrossEntropyLoss"),
        ("CustomLinearClsHead", "IBLoss"),
        ("CustomNonLinearClsHead", "IBLoss"),
    ]

    @pytest.fixture()
    def classifier(self, request, mocker, fxt_multi_class_cls_dataset_entity: DatasetEntity) -> SAMImageClassifier:
        head_type, loss_type = request.param
        n_data = len(fxt_multi_class_cls_dataset_entity)
        labels = fxt_multi_class_cls_dataset_entity.get_labels()
        num_classes = len(labels)
        cfg = {
            "backbone": None,
            "head": {
                "type": head_type,
                "num_classes": num_classes,
                "in_channels": 10,
                "loss": {
                    "type": loss_type,
                },
            },
            "multilabel": False,
            "hierarchical": False,
        }
        if loss_type == "IBLoss":
            cfg["head"]["loss"]["num_classes"] = num_classes
            cfg["head"]["loss"]["start"] = 0  # Should be zero

        class MockBackbone(torch.nn.Module):
            def forward(self, *args, **kwargs):
                return torch.randn([n_data, 10])

        mocker.patch("mmcls.models.classifiers.image.build_backbone", return_value=MockBackbone())
        classifier = SAMImageClassifier(track_loss_dynamics=True, **cfg)
        classifier.loss_dyns_tracker.init_with_otx_dataset(fxt_multi_class_cls_dataset_entity)
        return classifier

    @pytest.fixture()
    def data(self, fxt_multi_class_cls_dataset_entity: DatasetEntity):
        n_data = len(fxt_multi_class_cls_dataset_entity)
        labels = fxt_multi_class_cls_dataset_entity.get_labels()
        img = torch.rand(n_data, 3, 8, 8)
        gt_label = torch.arange(0, len(labels), dtype=torch.long).reshape(-1, 1)
        entity_ids = [item.id_ for item in fxt_multi_class_cls_dataset_entity]
        label_ids = [
            label.id_
            for item in fxt_multi_class_cls_dataset_entity
            for ann in item.get_annotations()
            for label in ann.get_labels()
        ]

        assert len(entity_ids) == len(label_ids)

        return {
            "img": img,
            "gt_label": gt_label,
            "img_metas": [
                {"entity_id": entity_id, "label_id": label_id} for entity_id, label_id in zip(entity_ids, label_ids)
            ],
        }

    @torch.no_grad()
    @pytest.mark.parametrize("classifier", TESTCASE, indirect=True, ids=lambda x: "-".join(x))
    def test_train_step(self, classifier: SAMImageClassifier, data: Dict[str, Any], tmp_dir_path: str):
        outputs = classifier.train_step(data)

        assert "loss_dyns" in outputs
        assert "entity_ids" in outputs
        assert "label_ids" in outputs

        n_steps = 3
        for iter in range(n_steps):
            classifier.loss_dyns_tracker.accumulate(outputs, iter)

        export_dir = osp.join(tmp_dir_path, "noisy_label_detection")
        classifier.loss_dyns_tracker.export(export_dir)

        dataset = dm.Dataset.import_from(export_dir, format="datumaro")

        for item in dataset:
            for ann in item.annotations:
                for k, v in ann.attributes.items():
                    assert k in {"iters", "loss_dynamics"}
                    assert len(v) == n_steps



"""Base task of OTX."""

# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
import os
import shutil
import tempfile
from abc import ABC, abstractmethod
from datetime import timedelta
from typing import Any, Dict, Iterable, List, Optional

import torch
from torch import distributed as dist

from otx.algorithms.common.adapters.mmcv.hooks import OTXLoggerHook
from otx.algorithms.common.adapters.mmcv.hooks.cancel_hook import CancelInterfaceHook
from otx.algorithms.common.configs.training_base import TrainType
from otx.algorithms.common.utils import UncopiableDefaultDict
from otx.algorithms.common.utils.logger import get_logger
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.explain_parameters import ExplainParameters
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label import LabelEntity
from otx.api.entities.metrics import MetricsGroup
from otx.api.entities.model import ModelEntity, ModelPrecision, OptimizationMethod
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.train_parameters import TrainParameters
from otx.api.serialization.label_mapper import LabelSchemaMapper
from otx.api.usecases.reporting.time_monitor_callback import TimeMonitorCallback
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.export_interface import ExportType, IExportTask
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.unload_interface import IUnload

TRAIN_TYPE_DIR_PATH = {
    TrainType.Incremental.name: ".",
    TrainType.Selfsupervised.name: "selfsl",
    TrainType.Semisupervised.name: "semisl",
}

logger = get_logger()


class OnHookInitialized:
    """OnHookInitialized class."""

    def __init__(self, task_instance):
        self.task_instance = task_instance
        self.__findable = False  # a barrier to block segmentation fault

    def __call__(self, cancel_interface):
        """Function call in OnHookInitialized."""
        if isinstance(self.task_instance, int) and self.__findable:
            import ctypes

            # NOTE: BE AWARE OF SEGMENTATION FAULT
            self.task_instance = ctypes.cast(self.task_instance, ctypes.py_object).value
        self.task_instance.cancel_hook_initialized(cancel_interface)

    def __repr__(self):
        """Function repr in OnHookInitialized."""
        return f"'{__name__}.OnHookInitialized'"

    def __deepcopy__(self, memo):
        """Function deepcopy in OnHookInitialized."""
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        result.task_instance = self.task_instance
        result.__findable = True  # pylint: disable=unused-private-member, protected-access
        return result

    def __reduce__(self):
        """Function reduce in OnHookInitialized."""
        return (self.__class__, (id(self.task_instance),))


# pylint: disable=too-many-instance-attributes
class OTXTask(IInferenceTask, IExportTask, IEvaluationTask, IUnload, ABC):
    """Base task of OTX."""

    def __init__(self, task_environment: TaskEnvironment, output_path: Optional[str] = None):
        self._config: Dict[Any, Any] = {}
        self._task_environment = task_environment
        self._task_type = task_environment.model_template.task_type
        self._labels = task_environment.get_labels(include_empty=False)
        self._work_dir_is_temp = False
        self._output_path = output_path
        self._output_path = output_path if output_path is not None else self._get_tmp_dir()
        self._time_monitor: Optional[TimeMonitorCallback] = None
        self.on_hook_initialized = OnHookInitialized(self)
        self._learning_curves = UncopiableDefaultDict(OTXLoggerHook.Curve)
        self._model_label_schema: List[LabelEntity] = []
        self._resume = False
        self._should_stop = False
        self.cancel_interface: Optional[CancelInterfaceHook] = None
        self.reserved_cancel = False
        self._model_ckpt = None
        self._precision = [ModelPrecision.FP32]
        self._optimization_methods: List[OptimizationMethod] = []
        self._is_training = False

        self.override_configs: Dict[str, str] = {}

        # This is for hpo, and this should be removed
        self.project_path = self._output_path

        if self._is_multi_gpu_training():
            self._setup_multigpu_training()

    @staticmethod
    def _is_multi_gpu_training():
        multi_gpu_env = ["MASTER_ADDR", "MASTER_PORT", "LOCAL_WORLD_SIZE", "WORLD_SIZE", "LOCAL_RANK", "RANK"]
        for env in multi_gpu_env:
            if env not in os.environ:
                return False

        return torch.cuda.is_available()

    @staticmethod
    def _setup_multigpu_training():
        if not dist.is_initialized():
            torch.cuda.set_device(int(os.environ["LOCAL_RANK"]))
            dist.init_process_group(backend="nccl", init_method="env://", timeout=timedelta(seconds=30))
            logger.info(f"Dist info: rank {dist.get_rank()} / {dist.get_world_size()} world_size")

    def _get_tmp_dir(self):
        self._work_dir_is_temp = True
        # If training is excuted with torchrun, set all trainings' output directory same
        if "TORCHELASTIC_RUN_ID" in os.environ:
            return os.path.join(tempfile.gettempdir(), f"OTX-task-torchelastic-{os.environ['TORCHELASTIC_RUN_ID']}")
        return tempfile.mkdtemp(prefix="OTX-task-")

    def _load_model(self):
        """Loading model from checkpoint."""

        def _load_model_label_schema(model: Optional[ModelEntity]):
            # If a model has been trained and saved for the task already, create empty model and load weights here
            if model and "label_schema.json" in model.model_adapters:
                import json

                buffer = json.loads(model.get_data("label_schema.json").decode("utf-8"))
                model_label_schema = LabelSchemaMapper().backward(buffer)
                return model_label_schema.get_labels(include_empty=False)
            return self._labels

        logger.info("loading the model from the task env.")
        model = self._task_environment.model
        state_dict = self._load_model_ckpt(model)
        if state_dict:
            self._model_ckpt = os.path.join(self._output_path, "env_model_ckpt.pth")
            if os.path.exists(self._model_ckpt):
                os.remove(self._model_ckpt)
            torch.save(state_dict, self._model_ckpt)
            self._model_label_schema = _load_model_label_schema(model)
            if model is not None:
                self._resume = model.model_adapters.get("resume", False)

    def _load_model_ckpt(self, model: Optional[ModelEntity]):
        if model and "weights.pth" in model.model_adapters:
            # If a model has been trained and saved for the task already, create empty model and load weights here
            buffer = io.BytesIO(model.get_data("weights.pth"))
            model_data = torch.load(buffer, map_location=torch.device("cpu"))
            return model_data
        return None

    @abstractmethod
    def train(
        self, dataset: DatasetEntity, output_model: ModelEntity, train_parameters: Optional[TrainParameters] = None
    ):
        """Train function for OTX task."""
        raise NotImplementedError

    @abstractmethod
    def infer(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ) -> DatasetEntity:
        """Main infer function."""
        raise NotImplementedError

    @abstractmethod
    def export(
        self,
        export_type: ExportType,
        output_model: ModelEntity,
        precision: ModelPrecision = ModelPrecision.FP32,
        dump_features: bool = True,
    ):
        """Export function of OTX Task."""
        raise NotImplementedError

    @abstractmethod
    def explain(
        self,
        dataset: DatasetEntity,
        explain_parameters: Optional[ExplainParameters] = None,
    ) -> DatasetEntity:
        """Main explain function of OTX Task."""
        raise NotImplementedError

    @abstractmethod
    def evaluate(
        self,
        output_resultset: ResultSetEntity,
        evaluation_metric: Optional[str] = None,
    ):
        """Evaluate function of OTX Task."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _generate_training_metrics(learning_curves, scores) -> Iterable[MetricsGroup[Any, Any]]:
        """Get Training metrics (epochs & scores).

        Parses the training logs to get metrics from the latest training run
        :return output List[MetricsGroup]
        """
        raise NotImplementedError

    @abstractmethod
    def save_model(self, output_model: ModelEntity):
        """Save best model weights in trining task."""
        raise NotImplementedError

    def cancel_training(self):
        """Cancel training function in trining task.

        Sends a cancel training signal to gracefully stop the optimizer. The signal consists of creating a
        '.stop_training' file in the current work_dir. The runner checks for this file periodically.
        The stopping mechanism allows stopping after each iteration, but validation will still be carried out. Stopping
        will therefore take some time.
        """
        logger.info("Cancel training requested.")
        self._should_stop = True
        if self.cancel_interface is not None:
            self.cancel_interface.cancel()
        else:
            logger.info("but training was not started yet. reserved it to cancel")
            self.reserved_cancel = True

    def cancel_hook_initialized(self, cancel_interface: CancelInterfaceHook):
        """Initialization of cancel_interface hook."""
        logger.info("cancel hook is initialized")
        self.cancel_interface = cancel_interface
        if self.reserved_cancel and self.cancel_interface:
            self.cancel_interface.cancel()

    def cleanup(self):
        """Clean up work directory if user specified it."""
        if self._work_dir_is_temp:
            self._delete_scratch_space()

    def _delete_scratch_space(self):
        """Remove model checkpoints and otx logs."""
        if os.path.exists(self._output_path):
            shutil.rmtree(self._output_path, ignore_errors=False)

    def unload(self):
        """Unload the task."""
        self.cleanup()
        if self._is_docker():
            logger.warning("Got unload request. Unloading models. Throwing Segmentation Fault on purpose")
            import ctypes

            ctypes.string_at(0)
        else:
            logger.warning("Got unload request, but not on Docker. Only clearing CUDA cache")
            torch.cuda.empty_cache()
            logger.warning(
                f"Done unloading. " f"Torch is still occupying {torch.cuda.memory_allocated()} bytes of GPU memory"
            )

    @staticmethod
    def _is_docker():
        """Checks whether the task runs in docker container.

        :return bool: True if task runs in docker
        """
        path = "/proc/self/cgroup"
        is_in_docker = False
        if os.path.isfile(path):
            with open(path, encoding="UTF-8") as f:
                is_in_docker = is_in_docker or any("docker" in line for line in f)
        is_in_docker = is_in_docker or os.path.exists("/.dockerenv")
        return is_in_docker

    @property
    def config(self):
        """Config of OTX task."""
        return self._config

    @config.setter
    def config(self, config: Dict[Any, Any]):
        self._config = config



"""Collections of IR-related utils for common OTX algorithms."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Any, Dict, Tuple

from openvino.runtime import Core, serialize


def embed_ir_model_data(xml_file: str, data_items: Dict[Tuple[str], Any]) -> None:
    """Embeds serialized data to IR xml file.

    Args:
        xml_file : a path to IR xml file.
        data_items : a dict with tuple-keyworded serialized objects.
    """

    core = Core()
    model = core.read_model(xml_file)
    for k, data in data_items.items():
        model.set_rt_info(data, list(k))
    serialize(model, xml_file)



import json
import shutil
import struct
from os import path as osp
from pathlib import Path

import cv2
import numpy as np
import pytest

from otx.api.entities.model import ModelOptimizationType
from otx.api.usecases.adapters.model_adapter import ModelAdapter
from otx.cli.utils import io as target_package
from otx.cli.utils.io import (
    get_explain_dataset_from_filelist,
    get_image_files,
    read_binary,
    read_label_schema,
    read_model,
    save_model_data,
    save_saliency_output,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit

IMG_DATA_FORMATS = (
    ".jpg",
    ".JPG",
    ".jpeg",
    ".JPEG",
    ".gif",
    ".GIF",
    ".bmp",
    ".BMP",
    ".tif",
    ".TIF",
    ".tiff",
    ".TIFF",
    ".png",
    ".PNG",
)


@e2e_pytest_unit
def test_save_model_data(mocker, tmp_dir):
    mock_model = mocker.MagicMock()
    model_adapter = ModelAdapter(b"fake")
    file_name = "model.pth"
    mock_model.model_adapters = {file_name: model_adapter}

    save_model_data(mock_model, tmp_dir)

    with open(osp.join(tmp_dir, file_name), "rb") as f:
        assert f.readline() == b"fake"


@e2e_pytest_unit
def test_read_binary(tmp_dir):
    file_path = osp.join(tmp_dir, "test.txt")
    with open(file_path, "w") as f:
        f.write("fake")

    assert read_binary(file_path) == b"fake"


@pytest.fixture
def model_adapter_keys():
    return (
        "confidence_threshold",
        "image_threshold",
        "pixel_threshold",
        "min",
        "max",
        "config.json",
    )


@e2e_pytest_unit
def test_read_xml_bin_model(mocker, model_adapter_keys, tmp_dir):
    # prepare
    tmp_dir = Path(tmp_dir)
    for key in model_adapter_keys:
        (tmp_dir / key).write_text(key)

    xml_model_path = tmp_dir / "model.xml"
    xml_model_path.write_text("xml_model")
    bin_model_path = tmp_dir / "model.bin"
    bin_model_path.write_text("bin_model")

    # run
    model = read_model(mocker.MagicMock(), str(xml_model_path), mocker.MagicMock())

    # check
    model_adapters = model.model_adapters
    assert model_adapters["openvino.xml"].data == b"xml_model"
    assert model_adapters["openvino.bin"].data == b"bin_model"
    for key in model_adapter_keys:
        assert model_adapters[key].data == bytes(key, "utf-8")


@e2e_pytest_unit
def test_read_pth_model(mocker, tmp_dir):
    # prepare
    tmp_dir = Path(tmp_dir)
    model_path = tmp_dir / "model.pth"
    model_path.write_text("model")
    mock_is_checkpoint_nncf = mocker.patch.object(target_package, "is_checkpoint_nncf")
    mock_is_checkpoint_nncf.return_value = True

    for i in range(1, 5):
        (tmp_dir / f"aux_model_{i}.pth").write_text(f"aux_{i}")

    # run
    model = read_model(mocker.MagicMock(), str(model_path), mocker.MagicMock())

    # check
    model_adapters = model.model_adapters
    assert model_adapters["weights.pth"].data == b"model"
    for i in range(1, 5):
        model_adapters[f"aux_model_{i}.pth"].data == bytes(f"aux_model_{i}.pth", "utf-8")
    assert model.optimization_type == ModelOptimizationType.NNCF


@pytest.fixture
def mock_zip_file(model_adapter_keys, tmp_dir) -> str:
    """Mock zip file.

    zip file structure is as below.
    model/model.xml
    model/model.bin
    model/config.json
    model_adapter_keys => should have float value
    """
    tmp_dir = Path(tmp_dir)
    model_zip_dir = tmp_dir / "model_zip"
    model_zip_dir.mkdir()
    model_dir_in_zip = model_zip_dir / "model"
    model_dir_in_zip.mkdir()
    (model_dir_in_zip / "model.xml").write_text("xml_model")
    (model_dir_in_zip / "model.bin").write_text("bin_model")
    with (model_dir_in_zip / "config.json").open("w") as f:
        model_config = {"model_parameters": {}}
        for key in model_adapter_keys:
            model_config["model_parameters"][key] = 0.1234
        model_config["model_parameters"]["labels"] = {"fake": "fake"}
        json.dump(model_config, f)

    shutil.make_archive(tmp_dir / "model", "zip", model_zip_dir)

    return tmp_dir / "model.zip"


@e2e_pytest_unit
def test_read_zip_model(mocker, model_adapter_keys, mock_zip_file):
    # run
    model = read_model(mocker.MagicMock(), str(mock_zip_file), mocker.MagicMock())

    # check
    model_adapters = model.model_adapters
    assert model_adapters["openvino.xml"].data == b"xml_model"
    assert model_adapters["openvino.bin"].data == b"bin_model"
    for key in model_adapter_keys:
        assert model_adapters[key].data == struct.pack("f", 0.1234)


@e2e_pytest_unit
def test_read_unknown_model(mocker):
    with pytest.raises(ValueError):
        read_model(mocker.MagicMock(), "fake.unknown", mocker.MagicMock())


@pytest.fixture
def mock_lebel_schema_mapper_instance(mocker):
    mock_lebel_schema_mapper_class = mocker.patch.object(target_package, "LabelSchemaMapper")
    mock_lebel_schema_mapper_instance = mocker.MagicMock()
    mock_lebel_schema_mapper_class.return_value = mock_lebel_schema_mapper_instance
    return mock_lebel_schema_mapper_instance


@e2e_pytest_unit
@pytest.mark.parametrize("extension", [".xml", ".bin", ".pth"])
def test_read_label_schema_with_model_file(mock_lebel_schema_mapper_instance, extension, tmp_dir):
    # prepare
    tmp_dir = Path(tmp_dir)

    model_path = tmp_dir / f"model.{extension}"
    model_path.write_text("fake")
    fake_label_schema = {"fake": "fake"}
    with (tmp_dir / "label_schema.json").open("w") as f:
        json.dump(fake_label_schema, f)

    # run
    read_label_schema(str(model_path))

    # check
    mock_lebel_schema_mapper_instance.backward.assert_called_once_with(fake_label_schema)


@e2e_pytest_unit
def test_read_label_schema_with_model_zipfile(mock_lebel_schema_mapper_instance, mock_zip_file):
    # run
    read_label_schema(str(mock_zip_file))

    # check
    mock_lebel_schema_mapper_instance.backward.assert_called_once_with({"fake": "fake"})


@e2e_pytest_unit
@pytest.mark.parametrize("img_data_format", IMG_DATA_FORMATS)
def test_get_single_image_files(img_data_format):
    img_files = get_image_files(f"fake_path/img{img_data_format}")

    assert img_files[0] == ("./", f"fake_path/img{img_data_format}")


@e2e_pytest_unit
def test_get_image_files_in_dir(tmp_dir):
    # prepare
    tmp_dir = Path(tmp_dir)
    sub_dir_1 = tmp_dir / "sub_dir_1"
    sub_dir_1_1 = sub_dir_1 / "sub_dir_1_1"
    sub_dir_2 = tmp_dir / "sub_dir_2"
    sub_dir_1_1.mkdir(parents=True)
    sub_dir_2.mkdir()
    (sub_dir_1_1 / "fake.jpg").write_text("fake")
    (sub_dir_2 / "fake.jpg").write_text("fake")

    # run
    img_files = get_image_files(str(tmp_dir))

    # check
    for img_file in img_files:
        assert img_file in ((str(sub_dir_1_1), "fake.jpg"), (str(sub_dir_2), "fake.jpg"))


@e2e_pytest_unit
def test_get_image_files_empty_dir(tmp_dir):
    assert get_image_files(tmp_dir) is None


@e2e_pytest_unit
@pytest.mark.parametrize(
    "process_saliency_maps",
    [True, False],
    ids=["w_post_processing", "wo_post_processing"],
)
def test_save_saliency_output(tmp_dir, process_saliency_maps):
    # prepare
    img = np.array([[100 for _ in range(3)] for _ in range(3)])
    saliency_map = np.zeros([3, 3], dtype=np.uint8)
    weight = 0.3

    # run
    save_saliency_output(process_saliency_maps, img, saliency_map, tmp_dir, "fake", weight=weight)

    # check
    if process_saliency_maps:
        saliency_map_file = Path(tmp_dir) / "fake_saliency_map.png"
    else:
        saliency_map_file = Path(tmp_dir) / "fake_saliency_map.tiff"
    assert saliency_map_file.exists()
    saved_saliency = cv2.imread(str(saliency_map_file))
    assert (saved_saliency == saliency_map).all()

    if process_saliency_maps:
        overlay_img = Path(tmp_dir) / "fake_overlay_img.png"
        assert overlay_img.exists()
        saved_overlay = cv2.imread(str(overlay_img))
        assert (saved_overlay == np.array([[100 * weight for _ in range(3)] for _ in range(3)])).all()


@e2e_pytest_unit
def test_get_explain_dataset_from_filelist(tmp_dir):
    # prepare
    tmp_dir = Path(tmp_dir)
    fake_img = np.array([[[i for _ in range(3)] for _ in range(3)] for i in range(1, 4)])
    cv2.imwrite(str(tmp_dir / "fake1.jpg"), fake_img)
    cv2.imwrite(str(tmp_dir / "fake2.jpg"), fake_img)
    image_files = [(tmp_dir, "fake1.jpg"), (tmp_dir, "fake2.jpg")]

    # run
    dataset_entity = get_explain_dataset_from_filelist(image_files)

    # check
    assert (dataset_entity[0].media.numpy == fake_img).all()
    assert (dataset_entity[1].media.numpy == fake_img).all()
